from __future__ import absolute_import, print_function, division, unicode_literals

import os
from abc import ABCMeta, abstractmethod
from six import add_metaclass

import numpy as np
from scipy.io import loadmat
from imageio import imsave
from scipy.ndimage import gaussian_filter, zoom

from tqdm import tqdm
from boltons.cacheutils import cached, LRU

from .roc import general_roc, general_rocs_per_positive
from .numba_utils import fill_fixation_map

from .utils import TemporaryDirectory, run_matlab_cmd, Cache, average_values, deprecated_class, remove_trailing_nans
from .datasets import Stimulus, Fixations
from .metrics import NSS
from .sampling_models import SamplingModelMixin


def handle_stimulus(stimulus):
    """
    Make sure that a stimulus is a `Stimulus`-object
    """
    if not isinstance(stimulus, Stimulus):
        stimulus = Stimulus(stimulus)
    return stimulus


def normalize_saliency_map(saliency_map, cdf, cdf_bins):
    """ Normalize saliency to make saliency values distributed according to a given CDF
    """

    smap = saliency_map.copy()
    shape = smap.shape
    smap = smap.flatten()
    smap = np.argsort(np.argsort(smap)).astype(float)
    smap /= 1.0*len(smap)

    inds = np.searchsorted(cdf, smap, side='right')
    smap = cdf_bins[inds]
    smap = smap.reshape(shape)
    smap = smap.reshape(shape)
    return smap


class FullShuffledNonfixationProvider(object):
    def __init__(self, stimuli, fixations, max_fixations_in_cache=500*1000*1000):
        self.stimuli = stimuli
        self.fixations = fixations
        cache_size = int(max_fixations_in_cache / len(self.fixations.x))
        self.cache = LRU(cache_size)
        self.nonfixations_for_image = cached(self.cache)(self._nonfixations_for_image)
        self.widths = np.asarray([s[1] for s in stimuli.sizes]).astype(float)
        self.heights = np.asarray([s[0] for s in stimuli.sizes]).astype(float)

    def _nonfixations_for_image(self, n):
        inds = ~(self.fixations.n == n)
        xs = (self.fixations.x[inds].copy()).astype(float)
        ys = (self.fixations.y[inds].copy()).astype(float)

        other_ns = self.fixations.n[inds]
        xs *= self.stimuli.sizes[n][1]/self.widths[other_ns]
        ys *= self.stimuli.sizes[n][0]/self.heights[other_ns]

        return xs.astype(int), ys.astype(int)

    def __call__(self, stimuli, fixations, i):
        assert stimuli is self.stimuli
        n = fixations.n[i]
        return self.nonfixations_for_image(n)


def _get_unfixated_values(saliency_map, ys, xs):
    """Return all saliency values that have not been fixated at leat once."""
    fixation_map = np.zeros(saliency_map.shape)
    fill_fixation_map(
        fixation_map,
        np.array([ys, xs]).T
    )
    return saliency_map[fixation_map == 0].flatten()


@add_metaclass(ABCMeta)
class ScanpathSaliencyMapModel(object):
    """
    Most general saliency model class. The model is neither
    assumed to be time-independet nor to be a probabilistic
    model.
    """

    @abstractmethod
    def conditional_saliency_map(self, stimulus, x_hist, y_hist, t_hist, attributes=None, out=None):
        """
        Return the models saliency map prediction depending on a fixation history
        for the n-th image.
        """
        raise NotImplementedError()

    def conditional_saliency_map_for_fixation(self, stimuli, fixations, fixation_index, out=None):
        return self.conditional_saliency_map(
            stimuli.stimulus_objects[fixations.n[fixation_index]],
            x_hist=remove_trailing_nans(fixations.x_hist[fixation_index]),
            y_hist=remove_trailing_nans(fixations.y_hist[fixation_index]),
            t_hist=remove_trailing_nans(fixations.t_hist[fixation_index]),
            attributes={key: getattr(fixations, key)[fixation_index] for key in fixations.__attributes__},
            out=out
        )

    def AUCs(self, stimuli, fixations, nonfixations='uniform', verbose=False):
        """
        Calulate AUC scores for fixations

        :type fixations : Fixations
        :param fixations : Fixation object to calculate the AUC scores for.

        :type nonfixations : string or Fixations
        :param nonfixations : Nonfixations to use for calculating AUC scores.
                              Possible values are:
                                  'uniform':  Use uniform nonfixation distribution (Judd-AUC), i.e.
                                              all pixels from the saliency map.
                                  'unfixated': Use all pixels from the saliency map except the fixated ones.
                                  'shuffled': Use all fixations from other images as nonfixations.
                                  fixations-object: For each image, use the fixations in this fixation
                                                    object as nonfixations

        :rtype : ndarray
        :return : list of AUC scores for each fixation,
                  ordered as in `fixations.x` (average=='fixation' or None)
                  or by image numbers (average=='image')
        """
        rocs_per_fixation = []
        rocs = {}
        out = None

        nonfix_ys = None
        nonfix_xs = None

        if isinstance(nonfixations, Fixations):
            nonfix_xs = []
            nonfix_ys = []
            for n in range(fixations.n.max()+1):
                inds = nonfixations.n == n
                nonfix_xs.append(nonfixations.x_int[inds].copy())
                nonfix_ys.append(nonfixations.y_int[inds].copy())

        if nonfixations == 'shuffled':
            nonfixations = FullShuffledNonfixationProvider(stimuli, fixations)

        for i in tqdm(range(len(fixations.x)), total=len(fixations.x), disable=not verbose):
            out = self.conditional_saliency_map_for_fixation(stimuli, fixations, i, out=out)
            positives = np.asarray([out[fixations.y_int[i], fixations.x_int[i]]])
            if nonfixations == 'uniform':
                negatives = out.flatten()
            elif nonfixations == 'unfixated':
                negatives = _get_unfixated_values(
                    out,
                    [fixations.y_int[i]], [fixations.x_int[i]]
                )
            elif nonfix_xs is not None:
                n = fixations.n[i]
                negatives = out[nonfix_ys[n], nonfix_xs[n]]
            elif callable(nonfixations):
                _nonfix_xs, _nonfix_ys = nonfixations(stimuli, fixations, i)
                negatives = out[_nonfix_ys.astype(int), _nonfix_xs.astype(int)]
            else:
                raise ValueError("Don't know how to handle nonfixations {}".format(nonfixations))

            positives = positives.astype(float)
            negatives = negatives.astype(float)

            this_roc, _, _ = general_roc(positives, negatives)
            rocs.setdefault(fixations.n[i], []).append(this_roc)
            rocs_per_fixation.append(this_roc)
        return np.asarray(rocs_per_fixation)

    def AUC(self, stimuli, fixations, nonfixations='uniform', average='fixation', verbose=False):
        """
        Calulate AUC scores for fixations

        :type fixations : Fixations
        :param fixations : Fixation object to calculate the AUC scores for.

        :type nonfixations : string or Fixations
        :param nonfixations : Nonfixations to use for calculating AUC scores.
                              Possible values are:
                                  'uniform':  Use uniform nonfixation distribution (Judd-AUC), i.e.
                                              all pixels from the saliency map.
                                  'unfixated': Use all pixels from the saliency map except the fixated ones.
                                  'shuffled': Use all fixations from other images as nonfixations.
                                  fixations-object: For each image, use the fixations in this fixation
                                                    object as nonfixations

        :type average : string
        :param average : How to average the AUC scores for each fixation.
                         Possible values are:
                             'image': average over images
                             'fixation' or None: Return AUC score for each fixation separately

        :rtype : ndarray
        :return : list of AUC scores for each fixation,
                  ordered as in `fixations.x` (average=='fixation' or None)
                  or by image numbers (average=='image')
        """
        aucs = self.AUCs(stimuli, fixations, nonfixations=nonfixations, verbose=verbose)
        return average_values(aucs, fixations, average=average)

    def sAUCs(self, stimuli, fixations, verbose=False):
        return self.AUCs(stimuli, fixations, nonfixations='shuffled', verbose=verbose)

    def sAUC(self, stimuli, fixations, average='fixation', verbose=False):
        return self.AUC(stimuli, fixations, nonfixations='shuffled', average=average, verbose=verbose)

    def NSSs(self, stimuli, fixations, verbose=False):
        values = np.empty(len(fixations.x))
        out = None

        for i in tqdm(range(len(fixations.x)), disable=not verbose, total=len(fixations.x)):
            out = self.conditional_saliency_map_for_fixation(stimuli, fixations, i, out=out)
            values[i] = NSS(out, fixations.x_int[i], fixations.y_int[i])
        return values

    def NSS(self, stimuli, fixations, average='fixation', verbose=False):
        nsss = self.NSSs(stimuli, fixations, verbose=verbose)
        return average_values(nsss, fixations, average=average)

    def set_params(self, **kwargs):
        """
        Set model parameters, if the model has parameters

        This method has to reset caches etc., if the depend on the parameters
        """
        if kwargs:
            raise ValueError('Unkown parameters!', kwargs)


class SaliencyMapModel(ScanpathSaliencyMapModel):
    """
    Most model class for saliency maps. The model is assumed
    to be stationary in time (i.e. all fixations are independent)
    but the model is not explicitly a probabilistic model.
    """

    def __init__(self, cache_location = None, caching=True,
                 memory_cache_size=None):
        self._cache = Cache(cache_location, memory_cache_size=memory_cache_size)
        self.caching = caching

    @property
    def cache_location(self):
        return self._cache.cache_location

    @cache_location.setter
    def cache_location(self, value):
        self._cache.cache_location = value

    def saliency_map(self, stimulus):
        """
        Get saliency map for given stimulus.

        To overwrite this function, overwrite `_saliency_map` as otherwise
        the caching mechanism is disabled.
        """
        stimulus = handle_stimulus(stimulus)
        if not self.caching:
            return self._saliency_map(stimulus.stimulus_data)
        stimulus_id = stimulus.stimulus_id
        if not stimulus_id in self._cache:
            self._cache[stimulus_id] = self._saliency_map(stimulus.stimulus_data)
        return self._cache[stimulus_id]

    @abstractmethod
    def _saliency_map(self, stimulus):
        """
        Overwrite this to implement you own SaliencyMapModel.

        Parameters
        ----------

        @type  stimulus: ndarray
        @param stimulus: stimulus for which the saliency map should be computed.
        """
        raise NotImplementedError()

    def conditional_saliency_map(self, stimulus, *args, **kwargs):
        return self.saliency_map(stimulus)

    def AUCs(self, stimuli, fixations, nonfixations='uniform', verbose=False):
        """
        Calulate AUC scores for fixations

        :type fixations : Fixations
        :param fixations : Fixation object to calculate the AUC scores for.

        :type nonfixations : string or Fixations
        :param nonfixations : Nonfixations to use for calculating AUC scores.
                              Possible values are:
                                  'uniform':  Use uniform nonfixation distribution (Judd-AUC), i.e.
                                              all pixels from the saliency map.
                                  'unfixated': Use all pixels from the saliency map except the fixated ones.
                                  'shuffled': Use all fixations from other images as nonfixations.
                                  fixations-object: For each image, use the fixations in this fixation
                                                    object as nonfixations

        :rtype : ndarray
        :return : list of AUC scores for each fixation,
                  ordered as in `fixations.x` (average=='fixation' or None)
                  or by image numbers (average=='image')
        """
        rocs_per_fixation = np.empty(len(fixations.x))

        nonfix_ys = None
        nonfix_xs = None

        if isinstance(nonfixations, Fixations):
            nonfix_xs = []
            nonfix_ys = []
            for n in range(fixations.n.max()+1):
                inds = nonfixations.n == n
                nonfix_xs.append(nonfixations.x_int[inds].copy())
                nonfix_ys.append(nonfixations.y_int[inds].copy())

        if nonfixations == 'shuffled':
            nonfixations = FullShuffledNonfixationProvider(stimuli, fixations)

        for n in tqdm(range(len(stimuli)), total=len(stimuli), disable = not verbose):
            inds = fixations.n == n
            if not inds.sum():
                continue
            out = self.saliency_map(stimuli.stimulus_objects[n])
            positives = np.asarray(out[fixations.y_int[inds], fixations.x_int[inds]])
            if nonfixations == 'uniform':
                negatives = out.flatten()
            elif nonfixations == 'unfixated':
                negatives = _get_unfixated_values(
                    out,
                    fixations.y_int[inds], fixations.x_int[inds]
                )
            elif nonfix_xs is not None:
                negatives = out[nonfix_ys[n], nonfix_xs[n]]
            elif callable(nonfixations):
                _nonfix_xs, _nonfix_ys = nonfixations(stimuli, fixations, np.nonzero(inds)[0][0])
                negatives = out[_nonfix_ys.astype(int), _nonfix_xs.astype(int)]
            else:
                raise TypeError("Cannot handle nonfixations {}".format(nonfixations))

            positives = positives.astype(float)
            negatives = negatives.astype(float)

            rocs = general_rocs_per_positive(positives, negatives)
            rocs_per_fixation[inds] = rocs

        return rocs_per_fixation

    def AUC_per_image(self, stimuli, fixations, nonfixations='uniform', thresholds='all', verbose=False):
        """
        Calulate AUC scores per image for fixations

        :type fixations : Fixations
        :param fixations : Fixation object to calculate the AUC scores for.

        :type nonfixations : string or Fixations
        :param nonfixations : Nonfixations to use for calculating AUC scores.
                              Possible values are:
                                  'uniform':  Use uniform nonfixation distribution (Judd-AUC), i.e.
                                              all pixels from the saliency map.
                                  'unfixated': Use all pixels from the saliency map except the fixated ones.
                                  'shuffled': Use all fixations from other images as nonfixations.
                                  fixations-object: For each image, use the fixations in this fixation
                                                    object as nonfixations

        :type thresholds: string, either of 'all' or 'fixations'
                          'all' uses all saliency values as threshold, computing the true performance of the saliency
                          map as a binary classifier on the given fixations and nonfixations
                          'fixations' uses only the fixated values as done in AUC_Judd.

        :rtype : ndarray
        :return : list of AUC scores for each image,
                  or by image numbers (average=='image')
        """
        rocs_per_image = []
        out = None

        nonfix_xs = None
        nonfix_ys = None

        if thresholds == 'all':
            judd = 0
        elif thresholds == 'fixations':
            judd = 1
        else:
            raise ValueError("Unknown value of `thresholds`: {}".format(thresholds))

        if isinstance(nonfixations, Fixations):
            nonfix_xs = []
            nonfix_ys = []
            for n in range(fixations.n.max()+1):
                inds = nonfixations.n == n
                nonfix_xs.append(nonfixations.x_int[inds].copy())
                nonfix_ys.append(nonfixations.y_int[inds].copy())

        if nonfixations == 'shuffled':
            nonfixations = FullShuffledNonfixationProvider(stimuli, fixations)

        for n in tqdm(range(len(stimuli)), disable=not verbose):
            out = self.saliency_map(stimuli.stimulus_objects[n])
            inds = fixations.n == n
            positives = np.asarray(out[fixations.y_int[inds], fixations.x_int[inds]])
            if nonfixations == 'uniform':
                negatives = out.flatten()
            elif nonfixations == 'unfixated':
                negatives = _get_unfixated_values(
                    out,
                    fixations.y_int[inds], fixations.x_int[inds]
                )
            elif nonfix_xs is not None:
                negatives = out[nonfix_ys[n], nonfix_xs[n]]
            elif callable(nonfixations):
                _nonfix_xs, _nonfix_ys = nonfixations(stimuli, fixations, np.nonzero(inds)[0][0])
                negatives = out[_nonfix_ys.astype(int), _nonfix_xs.astype(int)]
            else:
                raise TypeError("Cannot handle nonfixations {}".format(nonfixations))

            positives = positives.astype(float)
            negatives = negatives.astype(float)
            this_roc, _, _ = general_roc(positives, negatives, judd=judd)
            rocs_per_image.append(this_roc)
        return rocs_per_image

    def AUC(self, stimuli, fixations, nonfixations='uniform', average='fixation', thresholds='all', verbose=False):
        """
        Calulate AUC scores for fixations

        :type fixations : Fixations
        :param fixations : Fixation object to calculate the AUC scores for.

        :type nonfixations : string or Fixations
        :param nonfixations : Nonfixations to use for calculating AUC scores.
                              Possible values are:
                                  'uniform':  Use uniform nonfixation distribution (Judd-AUC), i.e.
                                              all pixels from the saliency map.
                                  'unfixated': Use all pixels from the saliency map except the fixated ones.
                                  'shuffled': Use all fixations from other images as nonfixations.
                                  fixations-object: For each image, use the fixations in this fixation
                                                    object as nonfixations

        :type average : string
        :param average : How to average the AUC scores for each fixation.
                         Possible values are:
                             'image': average over images
                             'fixation' or None: Return AUC score for each fixation separately

        :type thresholds: string, either of 'all' or 'fixations'
                          'all' uses all saliency values as threshold, computing the true performance of the saliency
                          map as a binary classifier on the given fixations and nonfixations
                          'fixations' uses only the fixated values as done in AUC_Judd.

        :rtype : ndarray
        :return : list of AUC scores for each fixation,
                  ordered as in `fixations.x` (average=='fixation' or None)
                  or by image numbers (average=='image')
        """
        if average not in ['fixation', 'image']:
            raise NotImplementedError()
        aucs = np.asarray(self.AUC_per_image(stimuli, fixations, nonfixations=nonfixations, thresholds=thresholds, verbose=verbose))
        if average == 'fixation':
            weights = np.zeros_like(aucs)
            for n in set(fixations.n):
                weights[n] = (fixations.n == n).mean()
            weights /= weights.sum()

            # take care of nans due to no fixations
            aucs[weights == 0] = 0

            return np.average(aucs, weights=weights)
        elif average == 'image':
            return np.mean(aucs)
        else:
            raise ValueError(average)

    def AUC_Judd(self, stimuli, fixations, jitter=True, noise_size=1.0/10000000, random_seed=42, verbose=False):
        if jitter:
            model = RandomNoiseSaliencyMapModel(
                self,
                noise_size=noise_size,
                random_seed=random_seed
            )
        else:
            model = self
        return model.AUC(
            stimuli,
            fixations,
            average='image',
            nonfixations='unfixated',
            thresholds='fixations',
            verbose=verbose
        )

    def fixation_based_KL_divergence(self, stimuli, fixations, nonfixations='shuffled', bins=10, eps=1e-20):
        """
        Calulate fixation-based KL-divergences for fixations

        :type fixations : Fixations
        :param fixations : Fixation object to calculate the AUC scores for.

        :type nonfixations : string or Fixations
        :param nonfixations : Nonfixations to use for calculating AUC scores.
                              Possible values are:
                                  'uniform':  Use uniform nonfixation distribution (Judd-AUC), i.e.
                                              all pixels from the saliency map.
                                  'shuffled': Use all fixations from other images as nonfixations.
                                  fixations-object: For each image, use the fixations in this fixation
                                                    object as nonfixations

        :type  bins : int
        :param bins : Number of bins to use in estimating the fixation based KL divergence

        :type  eps : float
        :param eps : regularization constant for the KL divergence to avoid logarithms of zero.


        :rtype : float
        :return : fixation based KL divergence
        """

        fixation_values = []
        nonfixation_values = []

        saliency_min = np.inf
        saliency_max = -np.inf

        for n in range(len(stimuli.stimuli)):
            saliency_map = self.saliency_map(stimuli.stimulus_objects[n])
            saliency_min = min(saliency_min, saliency_map.min())
            saliency_max = max(saliency_max, saliency_map.max())

            f = fixations[fixations.n == n]
            fixation_values.append(saliency_map[f.y_int, f.x_int])
            if nonfixations == 'uniform':
                nonfixation_values.append(saliency_map.flatten())
            elif nonfixations == 'shuffled':
                f = fixations[fixations.n != n]
                widths = np.asarray([s[1] for s in stimuli.sizes]).astype(float)
                heights = np.asarray([s[0] for s in stimuli.sizes]).astype(float)
                xs = (f.x.copy())
                ys = (f.y.copy())
                other_ns = f.n

                xs *= stimuli.sizes[n][1]/widths[other_ns]
                ys *= stimuli.sizes[n][0]/heights[other_ns]

                nonfixation_values.append(saliency_map[ys.astype(int), xs.astype(int)])
            else:
                nonfix = nonfixations[nonfixations.n == n]
                nonfixation_values.append(saliency_map[nonfix.y_int, nonfix.x_int])

        fixation_values = np.hstack(fixation_values)
        nonfixation_values = np.hstack(nonfixation_values)

        hist_range = saliency_min, saliency_max

        p_fix, _ = np.histogram(fixation_values, bins=bins, range=hist_range, density=True)
        p_fix += eps
        p_fix /= p_fix.sum()
        p_nonfix, _ = np.histogram(nonfixation_values, bins=bins, range=hist_range, density=True)
        p_nonfix += eps
        p_nonfix /= p_nonfix.sum()

        return (p_fix * (np.log(p_fix) - np.log(p_nonfix))).sum()

    def image_based_kl_divergences(self, stimuli, gold_standard, minimum_value=1e-20, log_regularization=0, quotient_regularization=0, convert_gold_standard=True, verbose=False):
        """Calculate image-based KL-Divergences between model and gold standard for each stimulus

        This metric computes the KL-Divergence between model predictions and a gold standard
        when interpreting these as fixation densities. As in the MIT saliency benchmark,
        saliency maps are interpreted as densities by dividing them by their summed value.

        To avoid problems with zeros, the minimum value is added to all saliency maps.
        Alternatively the kl divergence itself can be regularized (see Model.kl_divergences for details).

        If the gold standard is already a probabilistic model that should not be converted in a
        new (different!) probabilistic model, set `convert_gold_standard` to False.
        """
        def convert_model(model, minimum_value):
            from .models import Model

            class SimpleProbabilisticModel(Model):
                def __init__(self, model, minimum_value):
                    self.model = model
                    self.minimum_value = minimum_value
                    super(SimpleProbabilisticModel, self).__init__()

                def _normalize_saliency_map(self, smap):
                    if smap.min() < 0:
                        smap = smap - smap.min()
                    smap = smap + self.minimum_value

                    smap_sum = smap.sum()
                    if smap_sum:
                        smap = smap / smap_sum
                    else:
                        smap[:] = 1.0
                        smap /= smap.sum()

                    return smap

                def _log_density(self, stimulus):
                    smap = self._normalize_saliency_map(self.model.saliency_map(stimulus))
                    return np.log(smap)

            return SimpleProbabilisticModel(model, minimum_value)

        prob_model = convert_model(self, minimum_value)
        if convert_gold_standard:
            prob_gold_standard = convert_model(gold_standard, minimum_value)
        else:
            prob_gold_standard = gold_standard

        return prob_model.kl_divergences(
            stimuli,
            prob_gold_standard,
            log_regularization=log_regularization,
            quotient_regularization=quotient_regularization,
            verbose=verbose
        )

    def image_based_kl_divergence(self, stimuli, gold_standard, minimum_value=1e-20, log_regularization=0, quotient_regularization=0, convert_gold_standard=True, verbose=False):
        """Calculate image-based KL-Divergences between model and gold standard averaged over stimuli

        for more details, see `image_based_kl_divergences`.
        """
        return np.mean(self.image_based_kl_divergences(stimuli, gold_standard,
                                                       minimum_value=minimum_value,
                                                       convert_gold_standard=convert_gold_standard,
                                                       log_regularization=log_regularization,
                                                       quotient_regularization=quotient_regularization,
                                                       verbose=verbose))

    def KLDivs(self, *args, **kwargs):
        """Alias for image_based_kl_divergence"""
        return self.image_based_kl_divergences(*args, **kwargs)

    def KLDiv(self, *args, **kwargs):
        """Alias for image_based_kl_divergence"""
        return self.image_based_kl_divergence(*args, **kwargs)

    def CCs(self, stimuli, other, verbose=False):
        """ Calculate Correlation Coefficient Metric against some other model

        Returns performances for each stimulus. For performance over dataset,
        see `CC`
        """
        coeffs = []

        def normalize(saliency_map):
            saliency_map -= saliency_map.mean()
            std = saliency_map.std()

            if std:
                saliency_map /= std

            return saliency_map

        for s in tqdm(stimuli, disable=not verbose):
            smap1 = normalize(self.saliency_map(s).copy())
            smap2 = normalize(other.saliency_map(s).copy())

            coeffs.append(np.corrcoef(smap1.flatten(), smap2.flatten())[0, 1])
        return np.asarray(coeffs)

    def CC(self, stimuli, other, verbose=False):
        return self.CCs(stimuli, other, verbose=verbose).mean()

    def NSSs(self, stimuli, fixations, verbose=False):
        values = np.empty(len(fixations.x))
        for n, s in enumerate(tqdm(stimuli, disable=not verbose)):
            smap = self.saliency_map(s).copy()
            inds = fixations.n == n
            values[inds] = NSS(smap, fixations.x_int[inds], fixations.y_int[inds])

        return values

    def SIMs(self, stimuli, other, verbose=False):
        """ Calculate Similarity Metric against some other model

        Returns performances for each stimulus. For performance over dataset,
        see `SIM`
        """

        def _normalize_saliency_map(smap):
            if smap.min() < 0:
                smap = smap - smap.min()

            smap_sum = smap.sum()

            if smap_sum:
                smap = smap / smap_sum
            else:
                smap[:] = 1.0
                smap /= smap.sum()

            return smap

        values = []
        for s in tqdm(stimuli, disable=not verbose):
            smap1 = _normalize_saliency_map(self.saliency_map(s))

            smap2 = _normalize_saliency_map(other.saliency_map(s))

            values.append(np.min([smap1, smap2], axis=0).sum())
        return np.asarray(values)

    def SIM(self, stimuli, other, verbose=False):
        return self.SIMs(stimuli, other, verbose=verbose).mean()

    def __add__(self, other):
        if not isinstance(other, SaliencyMapModel):
            return NotImplemented

        return LambdaSaliencyMapModel([self, other], fn=lambda smaps: np.sum(smaps, axis=0, keepdims=False), caching=False)

    def __sub__(self, other):
        if not isinstance(other, SaliencyMapModel):
            return NotImplemented

        return LambdaSaliencyMapModel([self, other], fn=lambda smaps: smaps[0] - smaps[1], caching=False)

    def __mul__(self, other):
        if not isinstance(other, SaliencyMapModel):
            return NotImplemented

        return LambdaSaliencyMapModel([self, other], fn=lambda smaps: np.prod(smaps, axis=0, keepdims=False), caching=False)

    def __truediv__(self, other):
        if not isinstance(other, SaliencyMapModel):
            return NotImplemented

        return LambdaSaliencyMapModel([self, other], fn=lambda smaps: smaps[0] / smaps[1], caching=False)


class CachedSaliencyMapModel(SaliencyMapModel):
    """Saliency map model which uses only precached saliency maps
    """
    def __init__(self, cache_location, **kwargs):
        if cache_location is None:
            raise ValueError("CachedSaliencyMapModel needs a cache location!")
        super(CachedSaliencyMapModel, self).__init__(cache_location=cache_location, **kwargs)

    def _saliency_map(self, stimulus):
        raise NotImplementedError()


class MatlabSaliencyMapModel(SaliencyMapModel):
    """
    A model that creates it's saliency maps from a matlab script.

    The script has to take at least two arguments: The first argument
    will contain the filename which contains the stimulus (by default as png),
    the second argument contains the filename where the saliency map should be
    saved to (by default a .mat file). For more complicated scripts, you can
    overwrite the method `matlab_command`. It has to be a format string
    which takes the fields `stimulus` and `saliency_map` for the stimulus file
    and the saliency map file.
    """
    def __init__(self, script_file, stimulus_ext = '.png', saliency_map_ext='.mat', only_color_stimuli=False, **kwargs):
        """
        Initialize MatlabSaliencyModel

        Parameters
        ----------

        @type  script_file: string
        @param script_file: location of script file for Matlab/octave.
                            Matlab/octave will be run from this directory.

        @type  stimulus_ext: string, defaults to '.png'
        @param stimulus_ext: In which format the stimulus should be handed to the matlab script.

        @type  saliency_map_ext: string, defaults to '.png'
        @param saliency_map_ext: In which format the script will return the saliency map

        @type  only_color_stimuli: bool, defaults to `False`
        @param only_color_stimuli: If True, indicates that the script can handle only color stimuli.
                                   Grayscale stimuli will be converted to color stimuli by setting all
                                   RGB channels to the same value.
        """
        super(MatlabSaliencyMapModel, self).__init__(**kwargs)
        self.script_file = script_file
        self.stimulus_ext = stimulus_ext
        self.saliency_map_ext = saliency_map_ext
        self.only_color_stimuli = only_color_stimuli
        self.script_directory = os.path.dirname(script_file)
        script_name = os.path.basename(script_file)
        self.command, ext = os.path.splitext(script_name)

    def matlab_command(self, stimulus):
        """
        Construct the command to pass to matlab.

        Parameters
        ----------

        @type  stimulus: ndarray
        @param stimulus: The stimulus for which the saliency map should be generated.
                         In most cases, this argument should not be needed.

        @returns: string, the command to pass to matlab. The returned string has to be
                  a format string with placeholders for `stimulus` and `saliency_map`
                  where the files containing stimulus and saliency map will be inserted.
                  To change the type of these files, see the constructor.
        """
        return "{command}('{{stimulus}}', '{{saliency_map}}');".format(command=self.command)

    def _saliency_map(self, stimulus):
        with TemporaryDirectory(cleanup=True) as temp_dir:
            stimulus_file = os.path.join(temp_dir, 'stimulus'+self.stimulus_ext)
            if self.only_color_stimuli:
                if stimulus.ndim == 2:
                    new_stimulus = np.empty((stimulus.shape[0], stimulus.shape[1], 3), dtype=stimulus.dtype)
                    for i in range(3):
                        new_stimulus[:, :, i] = stimulus
                    stimulus = new_stimulus
            if self.stimulus_ext == '.png':
                imsave(stimulus_file, stimulus)
            else:
                raise ValueError(self.stimulus_ext)

            saliency_map_file = os.path.join(temp_dir, 'saliency_map'+self.saliency_map_ext)

            command = self.matlab_command(stimulus).format(stimulus=stimulus_file,
                                                           saliency_map=saliency_map_file)

            run_matlab_cmd(command, cwd = self.script_directory)

            if self.saliency_map_ext == '.mat':
                saliency_map = loadmat(saliency_map_file)['saliency_map']
            else:
                raise ValueError(self.saliency_map_ext)

            return saliency_map


class GaussianSaliencyMapModel(SaliencyMapModel):
    """Gaussian saliency map model with given width"""
    def __init__(self, width=0.5, **kwargs):
        super(GaussianSaliencyMapModel, self).__init__(**kwargs)
        self.width = width

    def _saliency_map(self, stimulus):
        height = stimulus.shape[0]
        width = stimulus.shape[1]
        YS, XS = np.mgrid[:height, :width].astype(float)
        XS /= width
        YS /= height
        XS -= 0.5
        YS -= 0.5
        r_squared = XS**2 + YS**2
        return np.ones((stimulus.shape[0], stimulus.shape[1]))*np.exp(-0.5*r_squared/(self.width)**2)


class FixationMap(SaliencyMapModel):
    """
    Fixation maps for given stimuli and fixations.

    With the keyword `kernel_size`, you can control whether
    the fixation map should be blured or just contain
    the actual fixations.

    If ignore_doublicates is True, multiple fixations in the same
    location will be counted as only one fixation (the fixation map
    won't have entries larger than 1).
    """
    def __init__(self, stimuli, fixations, kernel_size=None, convolution_mode='reflect', ignore_doublicates=False, *args, **kwargs):
        super(FixationMap, self).__init__(*args, **kwargs)

        self.xs = {}
        self.ys = {}
        for n in range(len(stimuli)):
            f = fixations[fixations.n == n]
            self.xs[stimuli.stimulus_ids[n]] = f.x.copy()
            self.ys[stimuli.stimulus_ids[n]] = f.y.copy()

        self.kernel_size = kernel_size
        self.convolution_mode = convolution_mode
        self.ignore_doublicates = ignore_doublicates

    def _saliency_map(self, stimulus):
        stimulus = Stimulus(stimulus)
        stimulus_id = stimulus.stimulus_id
        if stimulus.stimulus_id not in self.xs:
            raise ValueError('No Fixations known for this stimulus!')
        saliency_map = np.zeros(stimulus.size)
        ff = np.vstack([self.ys[stimulus_id].astype(int), self.xs[stimulus_id].astype(int)]).T
        fill_fixation_map(saliency_map, ff)

        if self.ignore_doublicates:
            saliency_map[saliency_map >= 1] = 1

        if self.kernel_size:
            saliency_map = gaussian_filter(saliency_map, self.kernel_size, mode=self.convolution_mode)
        return saliency_map


class ResizingSaliencyMapModel(SaliencyMapModel):
    def __init__(self, parent_model, verbose=True, **kwargs):
        if 'caching' not in kwargs:
            kwargs['caching'] = False
        super(ResizingSaliencyMapModel, self).__init__(**kwargs)
        self.parent_model = parent_model
        self.verbose = verbose

    def _saliency_map(self, stimulus):
        smap = self.parent_model.saliency_map(stimulus)

        target_shape = (stimulus.shape[0],
                        stimulus.shape[1])

        if smap.shape != target_shape:
            if self.verbose:
                print("Resizing saliency map", smap.shape, target_shape)
            x_factor = target_shape[1] / smap.shape[1]
            y_factor = target_shape[0] / smap.shape[0]

            smap = zoom(smap, [y_factor, x_factor], order=1, mode='nearest')

            assert smap.shape == target_shape

        return smap


class DisjointUnionMixin(object):
    def _split_fixations(self, stimuli, fixations):
        """ return list of [(inds, model)]
        """
        raise NotImplementedError()

    def eval_metric(self, metric_name, stimuli, fixations, **kwargs):
        result = np.empty(len(fixations.x))
        done = np.zeros_like(result).astype(bool)
        verbose = kwargs.get('verbose')
        for inds, model in tqdm(self._split_fixations(stimuli, fixations), disable = not verbose):
            assert done[inds].sum() == 0
            _f = fixations[inds]
            this_metric = getattr(model, metric_name)
            this_result = this_metric(stimuli, _f, **kwargs)
            result[inds] = this_result
            done[inds] = True
        assert all(done)
        return result


class DisjointUnionSaliencyMapModel(DisjointUnionMixin, ScanpathSaliencyMapModel):
    def AUCs(self, stimuli, fixations, **kwargs):
        return self.eval_metric('AUCs', stimuli, fixations, **kwargs)

    def AUC(self, stimuli, fixations, **kwargs):
        if kwargs.get('nonfixations', 'uniform') == 'shuffled':
            kwargs = dict(kwargs)
            kwargs['nonfixations'] = FullShuffledNonfixationProvider(stimuli, fixations)
        return super(DisjointUnionSaliencyMapModel, self).AUC(stimuli, fixations, **kwargs)

    def NSSs(self, stimuli, fixations, **kwargs):
        return self.eval_metric('NSSs', stimuli, fixations, **kwargs)


class SubjectDependentSaliencyMapModel(DisjointUnionSaliencyMapModel):
    def __init__(self, subject_models, **kwargs):
        super(SubjectDependentSaliencyMapModel, self).__init__(**kwargs)
        self.subject_models = subject_models

    def _split_fixations(self, stimuli, fixations):
        for s in self.subject_models:
            yield fixations.subjects == s, self.subject_models[s]

    def conditional_saliency_map(self, stimulus, x_hist, y_hist, t_hist, attributes=None, out=None, **kwargs):
        if 'subjects' not in attributes:
            raise ValueError("SubjectDependentSaliencyModel can't compute conditional saliency maps without subject indication!")
        return self.subject_models[attributes['subjects']].conditional_saliency_map(
            stimulus, x_hist, y_hist, t_hist, attributes=attributes, **kwargs)


class ExpSaliencyMapModel(SaliencyMapModel):
    def __init__(self, parent_model):
        super(ExpSaliencyMapModel, self).__init__(caching=False)
        self.parent_model = parent_model

    def _saliency_map(self, stimulus):
        return np.exp(self.parent_model.saliency_map(stimulus))


class BluringSaliencyMapModel(SaliencyMapModel):
    def __init__(self, parent_model, kernel_size, mode='nearest', **kwargs):
        super(BluringSaliencyMapModel, self).__init__(**kwargs)
        self.parent_model = parent_model
        self.kernel_size = kernel_size
        self.mode = mode

    def _saliency_map(self, stimulus):
        smap = self.parent_model.saliency_map(stimulus)
        smap = gaussian_filter(smap, self.kernel_size, mode=self.mode)
        return smap


class DigitizeMapModel(SaliencyMapModel):
    def __init__(self, parent_model, bins=256, return_ints=True):
        super(DigitizeMapModel, self).__init__(caching=False)
        self.parent_model = parent_model
        self.bins = bins
        self.return_ints = return_ints

    def _saliency_map(self, stimulus):
        smap = self.parent_model.saliency_map(stimulus)
        min = smap.min()
        max = smap.max()
        bins = np.linspace(min, max, num=self.bins+1)
        smap = np.digitize(smap, bins) - 1
        if self.return_ints:
            return smap
        else:
            return smap.astype(float)


class HistogramNormalizedSaliencyMapModel(SaliencyMapModel):
    def __init__(self, parent_model, histogram=None, **kwargs):
        super(HistogramNormalizedSaliencyMapModel, self).__init__(**kwargs)

        self.parent_model = parent_model

        if histogram is None:
            histogram = np.ones(256) / 256

        self.histogram = histogram
        self.histogram /= self.histogram.sum()
        self.bins = np.linspace(0, 1, len(self.histogram))
        self.cdf = np.cumsum(self.histogram)

    def _saliency_map(self, stimulus):
        smap = self.parent_model.saliency_map(stimulus)
        return normalize_saliency_map(smap, self.cdf, self.bins)


class LambdaSaliencyMapModel(SaliencyMapModel):
    """Applies a function to a list of saliency maps from other models"""
    def __init__(self, parent_models, fn, **kwargs):
        super(LambdaSaliencyMapModel, self).__init__(**kwargs)
        self.parent_models = parent_models
        self.fn = fn

    def _saliency_map(self, stimulus):
        saliency_maps = [model.saliency_map(stimulus) for model in self.parent_models]
        return self.fn(saliency_maps)


class RandomNoiseSaliencyMapModel(LambdaSaliencyMapModel):
    def __init__(self, parent_model, noise_size=1.0/10000000, random_seed=42, **kwargs):
        super(RandomNoiseSaliencyMapModel, self).__init__(
            [parent_model],
            self.add_jitter,
            **kwargs
        )
        self.rst = np.random.RandomState(seed=random_seed)
        self.noise_size = noise_size

    def add_jitter(self, saliency_maps):
        saliency_map = saliency_maps[0]
        return saliency_map + self.rst.randn(*saliency_map.shape)*self.noise_size


class DensitySaliencyMapModel(SaliencyMapModel):
    """Uses fixation density as predicted by a probabilistic model as saliency maps"""
    def __init__(self, parent_model, **kwargs):
        super(DensitySaliencyMapModel, self).__init__(caching=False, **kwargs)
        self.parent_model = parent_model

    def _saliency_map(self, stimulus):
        return np.exp(self.parent_model.log_density(stimulus))


class LogDensitySaliencyMapModel(SaliencyMapModel):
    """Uses fixation log density as predicted by a probabilistic model as saliency maps"""
    def __init__(self, parent_model, **kwargs):
        super(LogDensitySaliencyMapModel, self).__init__(caching=False, **kwargs)
        self.parent_model = parent_model

    def _saliency_map(self, stimulus):
        return self.parent_model.log_density(stimulus).copy()


class EqualizedSaliencyMapModel(SaliencyMapModel):
    """Equalizes saliency maps to have uniform histogram"""
    def __init__(self, parent_model, **kwargs):
        super(EqualizedSaliencyMapModel, self).__init__(caching=False, **kwargs)
        self.parent_model = parent_model

    def _saliency_map(self, stimulus):
        smap = self.parent_model.saliency_map(stimulus)
        smap = np.argsort(np.argsort(smap.flatten())).reshape(smap.shape)
        smap = smap.astype(float)
        smap /= np.prod(smap.shape)
        return smap


def nd_argmax(array):
    return np.unravel_index(np.argmax(array.flatten()), array.shape)


class WTASamplingMixin(SamplingModelMixin):
    def sample_fixation(self, stimulus, x_hist, y_hist, t_hist, attributes=None, verbose=False, rst=None):
        conditional_saliency_map = self.conditional_saliency_map(stimulus, x_hist, y_hist, t_hist, attributes=attributes)
        y, x = nd_argmax(conditional_saliency_map)

        if not t_hist:
            t = 0
        elif len(t_hist) == 1:
            t = t_hist[0] * 2
        else:
            t = t_hist[-1] + np.mean(np.diff(t_hist))

        return x, y, t


GeneralSaliencyMapModel = deprecated_class(deprecated_in='0.2.16', removed_in='1.0.0', details="Use ScanpathSaliencyMapModel instead")(ScanpathSaliencyMapModel)



from __future__ import absolute_import, print_function, division#, unicode_literals

import numpy as np

import theano
import theano.tensor as T
from theano.ifelse import ifelse


def nonlinearity(input, x, y, length):
    """
    Apply a pointwise nonlinearity to input

    The nonlinearity is a picewise linear function.
    The graph of the function is given by the vectors x and y.
    """
    parts = []
    for i in range(length-1):
        x1 = x[i]
        x2 = x[i+1]
        y1 = y[i]
        y2 = y[i+1]
        #print x1.tag
        part = (y2-y1)/(x2-x1)*(theano.tensor.clip(input, x1, x2)-x1)
        parts.append(part)
    output = y[0]
    for part in parts:
        output = output + part
    return output


def gaussian_filter(input, sigma, window_radius = 40):
    """
    Filter input with a Gaussian using mode `nearest`.

    input is expected to be three dimensional of type n times x times y
    """

    # Construction of 1d kernel
    #filter_1d = T.arange(-window_radius, window_radius+1)
    # Work around some strange theano bug
    filter_1d = T.arange(2*window_radius + 1) - window_radius

    filter_1d = T.exp(-0.5*filter_1d**2/sigma**2)
    filter_1d = filter_1d / filter_1d.sum()
    filter_1d = filter_1d.astype(input.dtype)

    W = filter_1d.dimshuffle([0, 'x'])
    W2 = filter_1d.dimshuffle(['x', 0])

    blur_input = input.dimshuffle(['x', 0, 1, 2])
    filter_W = W.dimshuffle(['x', 'x', 0, 1])
    filter_W2 = W2.dimshuffle(['x', 'x', 0, 1])

    # Construction of filter pipeline
    blur_input_start = blur_input[:, :, :1, :]
    blur_input_end = blur_input[:, :, -1:, :]
    padded_input = T.concatenate([blur_input_start]*window_radius+[blur_input]+[blur_input_end]*window_radius, axis=2)

    blur_op = T.nnet.conv2d(padded_input, filter_W, border_mode='valid', filter_shape=[1, 1, None, None])
    #x_min = (W.shape[1]-1)//2
    #x_max = input.shape[2]+(W.shape[1]-1)//2
    #y_min = (W.shape[0]-1)//2+window_radius
    #y_max = input.shape[1]+(W.shape[0]-1)//2+window_radius
    #cropped_output1 = blur_op[:, :, y_min:y_max, x_min:x_max]
    #cropped_output1_start = blur_op[:, :, y_min:y_max, x_min:x_min+1]
    #cropped_output1_end = blur_op[:, :, y_min:y_max, x_max-1:x_max]

    cropped_output1 = blur_op
    cropped_output1_start = blur_op[:, :, :, :1]
    cropped_output1_end = blur_op[:, :, :, -1:]
    padded_cropped_input = T.concatenate([cropped_output1_start]*window_radius
                                         + [cropped_output1]
                                         + [cropped_output1_end] * window_radius, axis=3)
    blur_op2 = T.nnet.conv2d(padded_cropped_input, filter_W2, border_mode='valid', filter_shape=[1, 1, None, None])
    #x_min2 = (W2.shape[1]-1)//2+window_radius
    #x_max2 = input.shape[2]+(W2.shape[1]-1)//2+window_radius
    #y_min2 = (W2.shape[0]-1)//2
    #y_max2 = input.shape[1]+(W2.shape[0]-1)//2
    cropped_output2 = blur_op2[0, :, :, :]  # [0, :, y_min2:y_max2, x_min2:x_max2]
    return cropped_output2


class Blur(object):
    def __init__(self, input, sigma=20.0, window_radius=60):
        self.input = input
        self.sigma = theano.shared(value=np.array(sigma, dtype=theano.config.floatX), name='sigma')
        apply_blur = T.gt(self.sigma, 0.0)
        no_blur = T.le(self.sigma, 0.0)
        self.output = ifelse(no_blur, input, gaussian_filter(input.dimshuffle('x', 0, 1), self.sigma, window_radius)[0, :, :])
        self.params = [self.sigma]


class Nonlinearity(object):
    def __init__(self, input, nonlinearity_ys = None):
        self.input = input
        #self.num_nonlinearity = num_nonlinearity
        if nonlinearity_ys is None:
            nonlinearity_ys = np.linspace(0, 1, num=20)
        nonlinearity_ys = nonlinearity_ys.astype(theano.config.floatX)
        self.nonlinearity_xs = theano.shared(value=np.linspace(0, 1, len(nonlinearity_ys)).astype(theano.config.floatX), name='nonlinearity_xs')
        self.nonlinearity_ys = theano.shared(value=nonlinearity_ys, name='nonlinearity_ys')
        self.output = nonlinearity(input, self.nonlinearity_xs, self.nonlinearity_ys, len(nonlinearity_ys))
        self.params = [self.nonlinearity_ys]


class LogNonlinearity(object):
    def __init__(self, input, nonlinearity_ys = None):
        self.input = input
        #self.num_nonlinearity = num_nonlinearity
        if nonlinearity_ys is None:
            nonlinearity_ys = np.linspace(0, 1, num=20)
        nonlinearity_ys = nonlinearity_ys.astype(theano.config.floatX)
        self.nonlinearity_xs = theano.shared(value=np.linspace(0, 1, len(nonlinearity_ys)).astype(theano.config.floatX), name='nonlinearity_xs')
        self.nonlinearity_ys = theano.shared(value=nonlinearity_ys, name='nonlinearity_ys')
        self.output = nonlinearity(input, self.nonlinearity_xs, T.exp(self.nonlinearity_ys), len(nonlinearity_ys))
        self.params = [self.nonlinearity_ys]


class CenterBias(object):
    def __init__(self, input, centerbias = None, alpha=1.0):
        self.input = input
        if centerbias is None:
            centerbias = np.ones(12)
        self.alpha = theano.shared(value = np.array(alpha).astype(theano.config.floatX), name='alpha')
        self.centerbias_ys = theano.shared(value=np.array(centerbias, dtype=theano.config.floatX), name='centerbias_ys')
        self.centerbias_xs = theano.shared(value=np.linspace(0, 1, len(centerbias), dtype=theano.config.floatX), name='centerbias_xs')

        height = T.cast(input.shape[0], theano.config.floatX)
        width = T.cast(input.shape[1], theano.config.floatX)
        x_coords = (T.arange(width) - 0.5*width) / (0.5*width)
        y_coords = (T.arange(height) - 0.5*height) / (0.5*height) + 0.0001  # We cannot have zeros in there because of grad

        x_coords = x_coords.dimshuffle('x', 0)
        y_coords = y_coords.dimshuffle(0, 'x')

        dists = T.sqrt(T.square(x_coords) + self.alpha*T.square(y_coords))
        self.max_dist = T.sqrt(1 + self.alpha)
        self.dists = dists/self.max_dist

        self.factors = nonlinearity(self.dists, self.centerbias_xs, self.centerbias_ys, len(centerbias))

        apply_centerbias = T.gt(self.centerbias_ys.shape[0], 2)
        self.output = ifelse(apply_centerbias, self.input*self.factors, self.input)
        self.params = [self.centerbias_ys, self.alpha]


class AdditiveCenterBias(object):
    def __init__(self, input, centerbias = None, alpha=1.0):
        self.input = input
        if centerbias is None:
            centerbias = np.ones(12)
        self.alpha = theano.shared(value = np.array(alpha).astype(theano.config.floatX), name='alpha')
        self.centerbias_ys = theano.shared(value=np.array(centerbias, dtype=theano.config.floatX), name='centerbias_ys')
        self.centerbias_xs = theano.shared(value=np.linspace(0, 1, len(centerbias), dtype=theano.config.floatX), name='centerbias_xs')

        height = T.cast(input.shape[0], theano.config.floatX)
        width = T.cast(input.shape[1], theano.config.floatX)
        x_coords = (T.arange(width) - 0.5*width) / (0.5*width)
        y_coords = (T.arange(height) - 0.5*height) / (0.5*height) + 0.0001  # We cannot have zeros in there because of grad

        x_coords = x_coords.dimshuffle('x', 0)
        y_coords = y_coords.dimshuffle(0, 'x')

        dists = T.sqrt(T.square(x_coords) + self.alpha*T.square(y_coords))
        self.max_dist = T.sqrt(1 + self.alpha)
        self.dists = dists/self.max_dist

        self.factors = nonlinearity(self.dists, self.centerbias_xs, self.centerbias_ys, len(centerbias))

        apply_centerbias = T.gt(self.centerbias_ys.shape[0], 2)
        self.output = ifelse(apply_centerbias, self.input+self.factors, self.input)
        self.params = [self.centerbias_ys, self.alpha]


class LogDensity(object):
    def __init__(self, input):
        self.input = input
        self.output = T.log(input / input.sum())


class LogDensityFromLogarithmicScale(object):
    def __init__(self, input):
        self.input = input
        self.output = input - T.log(T.exp(input).sum())


class AverageLogLikelihood(object):
    def __init__(self, log_densities, x_inds, y_inds):
        self.log_densities = log_densities
        self.log_likelihoods = log_densities[y_inds, x_inds]
        self.average_log_likelihood = self.log_likelihoods.mean()


class SaliencyMapProcessing(object):
    def __init__(self, saliency_map, x_inds = None, y_inds = None,
                 sigma = 0.0, window_radius = 80, nonlinearity_ys = None, centerbias = None, alpha = 1.0):

        self.saliency_map = saliency_map

        if x_inds is None:
            x_inds = T.lvector('x_inds')
        if y_inds is None:
            y_inds = T.lvector('y_inds')

        class TheanoObjects(object):
            pass

        self.theano_objects = TheanoObjects()

        self.x_inds = x_inds
        self.y_inds = y_inds
        self.theano_objects.blur = Blur(saliency_map, sigma=sigma, window_radius=window_radius)
        self.blur = self.theano_objects.blur.output

        self.theano_objects.nonlinearity = Nonlinearity(self.blur, nonlinearity_ys=nonlinearity_ys)
        self.nonlinearity = self.theano_objects.nonlinearity.output

        self.theano_objects.centerbias = CenterBias(self.nonlinearity, centerbias=centerbias, alpha=alpha)
        self.centerbias = self.theano_objects.centerbias.output

        self.theano_objects.log_density = LogDensity(self.centerbias)
        self.log_density = self.theano_objects.log_density.output

        self.theano_objects.average_log_likelihood = AverageLogLikelihood(self.log_density, self.x_inds, self.y_inds)

        self.average_log_likelihood = self.theano_objects.average_log_likelihood.average_log_likelihood
        self.params = self.theano_objects.blur.params + self.theano_objects.nonlinearity.params + self.theano_objects.centerbias.params

        self.blur_radius = self.theano_objects.blur.sigma
        self.nonlinearity_ys = self.theano_objects.nonlinearity.nonlinearity_ys
        self.centerbias_ys = self.theano_objects.centerbias.centerbias_ys
        self.alpha = self.theano_objects.centerbias.alpha


class SaliencyMapProcessingLogNonlinearity(object):
    def __init__(self, saliency_map, x_inds = None, y_inds = None,
                 sigma = 0.0, window_radius = 80, nonlinearity_ys = None, centerbias = None, alpha = 1.0):

        self.saliency_map = saliency_map

        if x_inds is None:
            x_inds = T.lvector('x_inds')
        if y_inds is None:
            y_inds = T.lvector('y_inds')

        class TheanoObjects(object):
            pass

        self.theano_objects = TheanoObjects()

        self.x_inds = x_inds
        self.y_inds = y_inds
        self.theano_objects.blur = Blur(saliency_map, sigma=sigma, window_radius=window_radius)
        self.blur = self.theano_objects.blur.output

        self.theano_objects.nonlinearity = LogNonlinearity(self.blur, nonlinearity_ys=nonlinearity_ys)
        self.nonlinearity = self.theano_objects.nonlinearity.output

        self.theano_objects.centerbias = CenterBias(self.nonlinearity, centerbias=centerbias, alpha=alpha)
        self.centerbias = self.theano_objects.centerbias.output

        self.theano_objects.log_density = LogDensity(self.centerbias)
        self.log_density = self.theano_objects.log_density.output

        self.theano_objects.average_log_likelihood = AverageLogLikelihood(self.log_density, self.x_inds, self.y_inds)

        self.average_log_likelihood = self.theano_objects.average_log_likelihood.average_log_likelihood
        self.params = self.theano_objects.blur.params + self.theano_objects.nonlinearity.params + self.theano_objects.centerbias.params

        self.blur_radius = self.theano_objects.blur.sigma
        self.nonlinearity_ys = self.theano_objects.nonlinearity.nonlinearity_ys
        self.centerbias_ys = self.theano_objects.centerbias.centerbias_ys
        self.alpha = self.theano_objects.centerbias.alpha


class SaliencyMapProcessingLogarithmic(object):
    def __init__(self, saliency_map, x_inds = None, y_inds = None,
                 sigma = 0.0, window_radius = 80, nonlinearity_ys = None, centerbias = None, alpha = 1.0):

        self.saliency_map = saliency_map

        if x_inds is None:
            x_inds = T.lvector('x_inds')
        if y_inds is None:
            y_inds = T.lvector('y_inds')

        class TheanoObjects(object):
            pass

        self.theano_objects = TheanoObjects()

        self.x_inds = x_inds
        self.y_inds = y_inds
        self.theano_objects.blur = Blur(saliency_map, sigma=sigma, window_radius=window_radius)
        self.blur = self.theano_objects.blur.output

        self.theano_objects.nonlinearity = Nonlinearity(self.blur, nonlinearity_ys=nonlinearity_ys)
        self.nonlinearity = self.theano_objects.nonlinearity.output

        self.theano_objects.centerbias = AdditiveCenterBias(self.nonlinearity, centerbias=centerbias, alpha=alpha)
        self.centerbias = self.theano_objects.centerbias.output

        self.theano_objects.log_density = LogDensityFromLogarithmicScale(self.centerbias)
        self.log_density = self.theano_objects.log_density.output

        self.theano_objects.average_log_likelihood = AverageLogLikelihood(self.log_density, self.x_inds, self.y_inds)

        self.average_log_likelihood = self.theano_objects.average_log_likelihood.average_log_likelihood
        self.params = self.theano_objects.blur.params + self.theano_objects.nonlinearity.params + self.theano_objects.centerbias.params

        self.blur_radius = self.theano_objects.blur.sigma
        self.nonlinearity_ys = self.theano_objects.nonlinearity.nonlinearity_ys
        self.centerbias_ys = self.theano_objects.centerbias.centerbias_ys
        self.alpha = self.theano_objects.centerbias.alpha



from __future__ import absolute_import, print_function, division

import zipfile
import os
import shutil
import warnings
import glob
import itertools
from six.moves import urllib

import numpy as np
from scipy.io import loadmat
from natsort import natsorted
import dill
from pkg_resources import resource_string
from PIL import Image
from boltons.fileutils import mkdir_p
from tqdm import tqdm

from .datasets import FileStimuli, Stimuli, FixationTrains, Fixations, read_hdf5
from .utils import TemporaryDirectory, filter_files, run_matlab_cmd, download_and_check, download_file_from_google_drive, check_file_hash
from .generics import progressinfo


def create_memory_stimuli(filenames):
    """
    Create a `Stimuli`-class from a list of filenames by reading the them
    """
    tmp_stimuli = FileStimuli(filenames)
    stimuli = list(tmp_stimuli.stimuli)  # Read all stimuli
    return Stimuli(stimuli)


def create_stimuli(stimuli_location, filenames, location=None):
    """
    Create a Stimuli class of stimuli.

    Parameters
    ----------

    @type  stimuli_location: string
    @param stimuli_location: the base path where the stimuli are located.
                             If `location` is provided, this directory will
                             be copied to `location` (see below).

    @type  filenames: list of strings
    @param filenames: lists the filenames of the stimuli to include in the dataset.
                      Filenames have to be relative to `stimuli_location`.

    @type  location: string or `None`
    @param location: If provided, the function will copy the filenames to
                     `location` and return a `FileStimuli`-object for the
                     copied files. Otherwise a `Stimuli`-object is returned.

    @returns: `Stimuli` or `FileStimuli` object depending on `location`.

    """
    if location is not None:
        shutil.copytree(stimuli_location,
                        location)
        filenames = [os.path.join(location, f) for f in filenames]

        return FileStimuli(filenames)

    else:
        filenames = [os.path.join(stimuli_location, f) for f in filenames]
        return create_memory_stimuli(filenames)


def _load(filename):
    """attempt to load hdf5 file and fallback to pickle files if present"""
    if os.path.isfile(filename):
        return read_hdf5(filename)

    stem, ext = os.path.splitext(filename)
    pydat_filename = stem+'.pydat'

    return dill.load(open(pydat_filename, 'rb'))


def get_toronto(location=None):
    """
    Loads or downloads and caches the Toronto dataset. The dataset
    consists of 120 color images of outdoor and indoor scenes
    of size 681x511px and the fixations of 20 subjects under
    free viewing conditions with 4 seconds presentation time.

    @type  location: string, defaults to `None`
    @param location: If and where to cache the dataset. The dataset
                     will be stored in the subdirectory `toronto` of
                     location and read from there, if already present.
    @return: Stimuli, FixationTrains

    .. warning::
        At the moment, the subjects stated in the FixationTrains object
        will not be correct, as they are difficult to infer from the
        published data (the data per subject is not stated in image dimensions)

    .. seealso::

        Neil Bruce, John K. Tsotsos. Attention based on information maximization [JoV 2007]

        `http://www-sop.inria.fr/members/Neil.Bruce/#SOURCECODE
    """
    if location:
        location = os.path.join(location, 'toronto')
        if os.path.exists(location):
            stimuli = _load(os.path.join(location, 'stimuli.hdf5'))
            fixations = _load(os.path.join(location, 'fixations.hdf5'))
            return stimuli, fixations
        os.makedirs(location)
    with TemporaryDirectory() as temp_dir:
        src = 'http://www-sop.inria.fr/members/Neil.Bruce/eyetrackingdata.zip'
        target = os.path.join(temp_dir, 'eyetrackingdata.zip')
        md5_sum = '38d5c02217060d4d2d1a4649cc632af1'
        download_and_check(src, target, md5_sum)
        z = zipfile.ZipFile(target)
        print('Extracting')
        z.extractall(temp_dir)

        # Stimuli
        stimuli_src_location = os.path.join(temp_dir, 'eyetrackingdata', 'fixdens', 'Original Image Set')
        stimuli_target_location = os.path.join(location, 'stimuli') if location else None
        stimuli_filenames = ['{}.jpg'.format(i) for i in range(1, 121)]

        stimuli = create_stimuli(stimuli_src_location, stimuli_filenames, stimuli_target_location)

        points = loadmat(os.path.join(temp_dir, 'eyetrackingdata', 'fixdens', 'origfixdata.mat'))['white']
        ts = []
        xs = []
        ys = []
        ns = []
        subjects = []
        for n in range(len(stimuli.stimuli)):
            _ys, _xs = np.nonzero(points[0, n])
            xs.extend([[x] for x in _xs])
            ys.extend([[y] for y in _ys])
            ns.extend([n for x in _xs])
            ts.extend([[0] for x in _xs])
            subjects.extend([0 for x in _xs])
        fixations = FixationTrains.from_fixation_trains(xs, ys, ts, ns, subjects)
    if location:
        stimuli.to_hdf5(os.path.join(location, 'stimuli.hdf5'))
        fixations.to_hdf5(os.path.join(location, 'fixations.hdf5'))
    return stimuli, fixations


def get_toronto_with_subjects(location=None):
    """
    Loads or downloads and caches the Toronto dataset. The dataset
    consists of 120 color images of outdoor and indoor scenes
    of size 681x511px and the fixations of 20 subjects under
    free viewing conditions with 4 seconds presentation time.

    @type  location: string, defaults to `None`
    @param location: If and where to cache the dataset. The dataset
                     will be stored in the subdirectory `toronto` of
                     location and read from there, if already present.
    @return: Stimuli, FixationTrains

    .. warning::
        This function uses the positions as given per subject in the toronto
        dataset. Unfortunately, these positions do not seem to be pixel positions
        as the fixations are not located as in the file `origfixdata.mat'.
        The code is still in the package as a template for fixing this issue.

    .. seealso::

        Neil Bruce, John K. Tsotsos. Attention based on information maximization [JoV 2007]

        `http://www-sop.inria.fr/members/Neil.Bruce/#SOURCECODE
    """
    warnings.warn('This function reports wrong fixation positions! See docstring for details.')
    if location:
        location = os.path.join(location, 'toronto')
        os.makedirs(location)
    with TemporaryDirectory() as temp_dir:
        src = 'http://www-sop.inria.fr/members/Neil.Bruce/eyetrackingdata.zip'
        target = os.path.join(temp_dir, 'eyetrackingdata.zip')
        md5_sum = '38d5c02217060d4d2d1a4649cc632af1'
        download_and_check(src, target, md5_sum)
        z = zipfile.ZipFile(target)
        print('Extracting')
        z.extractall(temp_dir)

        # Stimuli
        stimuli_src_location = os.path.join(temp_dir, 'eyetrackingdata', 'fixdens', 'Original Image Set')
        stimuli_target_location = os.path.join(location, 'stimuli') if location else None
        stimuli_filenames = ['{}.jpg'.format(i) for i in range(1, 121)]

        stimuli = create_stimuli(stimuli_src_location, stimuli_filenames, stimuli_target_location)

        print("Getting fixations")
        raw_path = os.path.join(temp_dir, 'eyetrackingdata', 'fixdens', 'Raw')
        subjects = os.listdir(raw_path)
        train_xs = []
        train_ys = []
        train_ts = []
        train_ns = []
        train_subjects = []
        subjects = [s for s in subjects if s != 'processed']
        subjects = sorted(subjects)
        for subject_nr, subject in enumerate(subjects):
            print("Doing subject", subject)
            for n, image in enumerate(stimuli_filenames):
                imagename = os.path.splitext(os.path.split(image)[-1])[0]
                content = open(os.path.join(raw_path, subject, imagename+'.fix')).read()
                content = content.replace('\r', '')
                if 'No fixations' in content:
                    print("No fixations for {}, skipping".format(image))
                    continue
                subject_fixations = content.split('Fixation Listing')[1].split('\n\n')[1].split('Average')[0]
                _xs = []
                _ys = []
                _ts = []
                for line in subject_fixations.split('\n'):
                    if not line:
                        continue
                    parts = line.split()
                    parts = [p.replace(',', '') for p in parts]
                    _xs.append(int(parts[1]))
                    _ys.append(int(parts[2]))
                    _ts.append(float(parts[3]))
                _xs = np.array(_xs, dtype=np.float)
                _ys = np.array(_ys, dtype=np.float)
                _ts = np.array(_ts, dtype=np.float)
                xs = []
                ys = []
                ts = []
                for i in range(len(_xs)):
                    if _xs[i] > 680:
                        continue
                    if _ys[i] > 510:
                        continue
                    xs.append(_xs[i])
                    ys.append(_ys[i])
                    ts.append(_ts[i])
                train_xs.append(xs)
                train_ys.append(ys)
                train_ts.append(ts)
                train_ns.append(n)
                train_subjects.append(subject_nr)

        fixations = FixationTrains.from_fixation_trains(train_xs, train_ys, train_ts, train_ns, train_subjects)

    if location:
        stimuli.to_hdf5(os.path.join(location, 'stimuli.hdf5'))
        fixations.to_hdf5(os.path.join(location, 'fixations.hdf5'))
    return stimuli, fixations


def _get_mit1003(dataset_name, location=None, include_initial_fixation=False, only_1024_by_768=False):
    """
    .. seealso::

        Tilke Judd, Krista Ehinger, Fredo Durand, Antonio Torralba. Learning to Predict where Humans Look [ICCV 2009]

        http://people.csail.mit.edu/tjudd/WherePeopleLook/index.html
    """

    if include_initial_fixation:
        first_fixation = 0
    else:
        first_fixation = 1

    if location:
        location = os.path.join(location, dataset_name)
        if os.path.exists(location):
            stimuli = _load(os.path.join(location, 'stimuli.hdf5'))
            fixations = _load(os.path.join(location, 'fixations.hdf5'))
            return stimuli, fixations
        os.makedirs(location)
    with TemporaryDirectory(cleanup=True) as temp_dir:
        download_and_check('http://people.csail.mit.edu/tjudd/WherePeopleLook/ALLSTIMULI.zip',
                           os.path.join(temp_dir, 'ALLSTIMULI.zip'),
                           '0d7df8b954ecba69b6796e77b9afe4b6')
        download_and_check('http://people.csail.mit.edu/tjudd/WherePeopleLook/DATA.zip',
                           os.path.join(temp_dir, 'DATA.zip'),
                           'ea19d74ad0a0144428c53e9d75c2d71c')
        download_and_check('http://people.csail.mit.edu/tjudd/WherePeopleLook/Code/DatabaseCode.zip',
                           os.path.join(temp_dir, 'DatabaseCode.zip'),
                           'd8e5e2b6ec827f4115ddbff59b0bdf1d')

        # Stimuli
        print('Creating stimuli')
        f = zipfile.ZipFile(os.path.join(temp_dir, 'ALLSTIMULI.zip'))
        namelist = f.namelist()
        namelist = filter_files(namelist, ['.svn', '__MACOSX', '.DS_Store'])
        f.extractall(temp_dir, namelist)

        stimuli_src_location = os.path.join(temp_dir, 'ALLSTIMULI')
        stimuli_target_location = os.path.join(location, 'stimuli') if location else None
        images = glob.glob(os.path.join(stimuli_src_location, '*.jpeg'))
        images = [os.path.split(img)[1] for img in images]
        stimuli_filenames = natsorted(images)

        if only_1024_by_768:
            def check_size(f):
                img = Image.open(os.path.join(stimuli_src_location, f))
                return img.size == (1024, 768)

            stimuli_filenames = [s for s in stimuli_filenames if check_size(s)]

        stimuli = create_stimuli(stimuli_src_location, stimuli_filenames, stimuli_target_location)

        # FixationTrains

        print('Creating fixations')
        f = zipfile.ZipFile(os.path.join(temp_dir, 'DATA.zip'))
        namelist = f.namelist()
        namelist = filter_files(namelist, ['.svn', '__MACOSX', '.DS_Store'])
        f.extractall(temp_dir, namelist)

        f = zipfile.ZipFile(os.path.join(temp_dir, 'DatabaseCode.zip'))
        namelist = f.namelist()
        namelist = filter_files(namelist, ['.svn', '__MACOSX', '.DS_Store'])
        f.extractall(temp_dir, namelist)

        subjects = glob.glob(os.path.join(temp_dir, 'DATA', '*'))
        # Exclude files
        subjects = [s for s in subjects if not os.path.splitext(s)[1]]
        subjects = [os.path.basename(s) for s in subjects]
        subjects = sorted(subjects)

        with open(os.path.join(temp_dir, 'extract_fixations.m'), 'wb') as f:
            f.write(resource_string(__name__, 'scripts/{}'.format('extract_fixations.m')))

        cmds = []
        # It is vital _not_ to store the extracted fixations in the main
        # directory where matlab is running, as matlab will check the timestamp
        # of all files in this directory very often. This leads to heavy
        # performance penalties and would make matlab run for more than an
        # hour.
        out_path = 'extracted'
        os.makedirs(os.path.join(temp_dir, out_path))
        total_cmd_count = len(stimuli_filenames) * len(subjects)
        for n, stimulus in enumerate(stimuli_filenames):
            for subject_id, subject in enumerate(subjects):
                subject_path = os.path.join('DATA', subject)
                outfile = '{0}_{1}.mat'.format(stimulus, subject)
                outfile = os.path.join(out_path, outfile)
                cmds.append("fprintf('%d/%d\\n', {}, {});".format(n*len(subjects)+subject_id, total_cmd_count))
                cmds.append("extract_fixations('{0}', '{1}', '{2}');".format(stimulus, subject_path, outfile))

        print('Running original code to extract fixations. This can take some minutes.')
        print('Warning: In the IPython Notebook, the output is shown on the console instead of the notebook.')
        with open(os.path.join(temp_dir, 'extract_all_fixations.m'), 'w') as f:
            for cmd in cmds:
                f.write('{}\n'.format(cmd))

        run_matlab_cmd('extract_all_fixations;', cwd=temp_dir)
        xs = []
        ys = []
        ts = []
        ns = []
        train_subjects = []
        for n, stimulus in enumerate(stimuli_filenames):
            stimulus_size = stimuli.sizes[n]
            for subject_id, subject in enumerate(subjects):
                subject_name = os.path.split(subject)[-1]
                outfile = '{0}_{1}.mat'.format(stimulus, subject_name)
                mat_data = loadmat(os.path.join(temp_dir, out_path, outfile))
                fix_data = mat_data['fixations']
                starts = mat_data['starts']
                x = []
                y = []
                t = []
                # if first_fixation == 1 the first fixation is skipped, as done
                # by Judd.
                for i in range(first_fixation, fix_data.shape[0]):
                    if fix_data[i, 0] < 0 or fix_data[i, 1] < 0:
                        continue
                    if fix_data[i, 0] >= stimulus_size[1] or fix_data[i, 1] >= stimulus_size[0]:
                        continue
                    x.append(fix_data[i, 0])
                    y.append(fix_data[i, 1])
                    t.append(starts[0, i]/240.0)  # Eye Tracker rate = 240Hz
                xs.append(x)
                ys.append(y)
                ts.append(t)
                ns.append(n)
                train_subjects.append(subject_id)
        fixations = FixationTrains.from_fixation_trains(xs, ys, ts, ns, train_subjects)

    if location:
        stimuli.to_hdf5(os.path.join(location, 'stimuli.hdf5'))
        fixations.to_hdf5(os.path.join(location, 'fixations.hdf5'))
    return stimuli, fixations


def get_mit1003(location=None):
    """
    Loads or downloads and caches the MIT1003 dataset. The dataset
    consists of 1003 natural indoor and outdoor scenes of
    sizes: max dim: 1024px, other dim: 405-1024px
    and the fixations of 15 subjects under
    free viewing conditions with 3 seconds presentation time.

    All fixations outside of the image are discarded. This includes
    blinks.

    @type  location: string, defaults to `None`
    @param location: If and where to cache the dataset. The dataset
                     will be stored in the subdirectory `toronto` of
                     location and read from there, if already present.
    @return: Stimuli, FixationTrains

    .. note::
        This code needs a working matlab or octave installation as the original
        matlab code by Judd et al. is used to extract the fixation from the
        eyetracking data.

        The first fixation of each fixation train is discarded as stated in the
        paper (Judd et al. 2009).

    .. seealso::

        Tilke Judd, Krista Ehinger, Fredo Durand, Antonio Torralba. Learning to Predict where Humans Look [ICCV 2009]

        http://people.csail.mit.edu/tjudd/WherePeopleLook/index.html
    """
    return _get_mit1003('MIT1003', location=location, include_initial_fixation=False)


def get_mit1003_with_initial_fixation(location=None):
    """
    Loads or downloads and caches the MIT1003 dataset. The dataset
    consists of 1003 natural indoor and outdoor scenes of
    sizes: max dim: 1024px, other dim: 405-1024px
    and the fixations of 15 subjects under
    free viewing conditions with 3 seconds presentation time.

    All fixations outside of the image are discarded. This includes
    blinks.

    @type  location: string, defaults to `None`
    @param location: If and where to cache the dataset. The dataset
                     will be stored in the subdirectory `toronto` of
                     location and read from there, if already present.
    @return: Stimuli, FixationTrains

    .. note::
        This code needs a working matlab or octave installation as the original
        matlab code by Judd et al. is used to extract the fixation from the
        eyetracking data.

        Unlike in the original paper (Judd et al. 2009), in this version of
        the MIT1003 dataset the first fixation of each scanpath
        is *not* discarded. See `get_mit1003` if you want to use
        the dataset as suggested by Judd.

    .. seealso::

        Tilke Judd, Krista Ehinger, Fredo Durand, Antonio Torralba. Learning to Predict where Humans Look [ICCV 2009]

        http://people.csail.mit.edu/tjudd/WherePeopleLook/index.html
    """
    return _get_mit1003('MIT1003_initial_fix', location=location, include_initial_fixation=True)


def get_mit1003_onesize(location=None):
    """
    Loads or downloads and caches the subset of the MIT1003 dataset
    used in "How close are we to understanding image-based
    saliency" (http://arxiv.org/abs/1409.7686) and
    "Deep Gaze I: Boosting Saliency Prediction with Feature Maps Trained
    on ImageNet" (http://arxiv.org/abs/1411.1045).
    The dataset conists of 463 natural indoor and outdoor scenes of
    size 1024 x 768 px and the fixations of 15 subjects under
    free viewing conditions with 3 seconds presentation time.

    All fixations outside of the image are discarded. This includes
    blinks.

    @type  location: string, defaults to `None`
    @param location: If and where to cache the dataset. The dataset
                     will be stored in the subdirectory `toronto` of
                     location and read from there, if already present.
    @return: Stimuli, FixationTrains

    .. note::
        This code needs a working matlab or octave installation as the original
        matlab code by Judd et al. is used to extract the fixation from the
        eyetracking data.

        The first fixation of each fixation train is discarded as stated in the
        paper (Judd et al. 2009).

    .. seealso::

        Tilke Judd, Krista Ehinger, Fredo Durand, Antonio Torralba. Learning to Predict where Humans Look [ICCV 2009]

        http://people.csail.mit.edu/tjudd/WherePeopleLook/index.html
    """
    return _get_mit1003('MIT1003_onesize', location=location, include_initial_fixation=False, only_1024_by_768=True)


def get_mit300(location=None):
    """
    Loads or downloads and caches the MIT300 test stimuli for
    the MIT saliency benchmark. The dataset
    consists of 300 stimuli
    and the fixations of 40 subjects under
    free viewing conditions with 3 seconds presentation time.

    @type  location: string, defaults to `None`
    @param location: If and where to cache the dataset. The dataset
                     will be stored in the subdirectory `toronto` of
                     location and read from there, if already present.
    @return: Stimuli

    .. seealso::

        Zoya Bylinskii and Tilke Judd and Ali Borji and Laurent Itti and Fr{\'e}do Durand and Aude Oliva and Antonio Torralba. MIT Saliency Benchmark

        http://saliency.mit.edu
    """
    if location:
        location = os.path.join(location, 'MIT300')
        if os.path.exists(location):
            stimuli = _load(os.path.join(location, 'stimuli.hdf5'))
            return stimuli
        os.makedirs(location)
    with TemporaryDirectory(cleanup=True) as temp_dir:
        download_and_check('http://saliency.mit.edu/BenchmarkIMAGES.zip',
                           os.path.join(temp_dir, 'BenchmarkIMAGES.zip'),
                           '03ed32bdf5e4289950cd28df89451260')

        # Stimuli
        print('Creating stimuli')
        f = zipfile.ZipFile(os.path.join(temp_dir, 'BenchmarkIMAGES.zip'))
        namelist = f.namelist()
        namelist = filter_files(namelist, ['.svn', '__MACOSX', '.DS_Store'])
        f.extractall(temp_dir, namelist)

        stimuli_src_location = os.path.join(temp_dir, 'BenchmarkIMAGES')
        stimuli_target_location = os.path.join(location, 'stimuli') if location else None
        images = glob.glob(os.path.join(stimuli_src_location, '*.jpg'))
        images = [os.path.split(img)[1] for img in images]
        stimuli_filenames = natsorted(images)

        stimuli = create_stimuli(stimuli_src_location, stimuli_filenames, stimuli_target_location)

    if location:
        stimuli.to_hdf5(os.path.join(location, 'stimuli.hdf5'))
    return stimuli


def get_cat2000_test(location=None):
    """
    Loads or downloads and caches the CAT2000 test dataset. The dataset
    consists of 1003 natural indoor and outdoor scenes of
    sizes: max dim: 1024px, other dim: 405-1024px
    and the fixations of 15 subjects under
    free viewing conditions with 3 seconds presentation time.

    All fixations outside of the image are discarded. This includes
    blinks.

    @type  location: string, defaults to `None`
    @param location: If and where to cache the dataset. The dataset
                     will be stored in the subdirectory `toronto` of
                     location and read from there, if already present.
    @return: Stimuli, FixationTrains

    .. note::
        This code needs a working matlab or octave installation as the original
        matlab code by Judd et al. is used to extract the fixation from the
        eyetracking data.

        The first fixation of each fixation train is discarded as stated in the
        paper (Judd et al. 2009).

    .. seealso::

        Tilke Judd, Krista Ehinger, Fredo Durand, Antonio Torralba. Learning to Predict where Humans Look [ICCV 2009]

        http://people.csail.mit.edu/tjudd/WherePeopleLook/index.html
    """
    if location:
        location = os.path.join(location, 'CAT2000_test')
        if os.path.exists(location):
            stimuli = _load(os.path.join(location, 'stimuli.hdf5'))
            return stimuli
        os.makedirs(location)
    with TemporaryDirectory(cleanup=True) as temp_dir:
        download_and_check('http://saliency.mit.edu/testSet.zip',
                           os.path.join(temp_dir, 'testSet.zip'),
                           'bd2ad31a0c4f02616b735eed2051b903')

        # Stimuli
        print('Creating stimuli')
        f = zipfile.ZipFile(os.path.join(temp_dir, 'testSet.zip'))
        f.extractall(temp_dir)

        stimuli_src_location = os.path.join(temp_dir, 'Stimuli')
        stimuli_target_location = os.path.join(location, 'Stimuli') if location else None
        images = glob.glob(os.path.join(stimuli_src_location, '**', '*.jpg'))
        images = [os.path.relpath(img, start=stimuli_src_location) for img in images]
        stimuli_filenames = natsorted(images)

        stimuli = create_stimuli(stimuli_src_location, stimuli_filenames, stimuli_target_location)

    if location:
        stimuli.to_hdf5(os.path.join(location, 'stimuli.hdf5'))
    return stimuli


def get_cat2000_train(location=None, include_initial_fixation=True):
    name = 'CAT2000_train'

    if not include_initial_fixation:
        name += '_without_initial_fixation'

    return _get_cat2000_train(name=name, location=location, include_initial_fixation=include_initial_fixation)


def _get_cat2000_train(name, location, include_initial_fixation):
    """
    Loads or downloads and caches the CAT2000 dataset. The dataset
    consists of 1003 natural indoor and outdoor scenes of
    sizes: max dim: 1024px, other dim: 405-1024px
    and the fixations of 15 subjects under
    free viewing conditions with 3 seconds presentation time.

    All fixations outside of the image are discarded. This includes
    blinks.

    @type  location: string, defaults to `None`
    @param location: If and where to cache the dataset. The dataset
                     will be stored in the subdirectory `toronto` of
                     location and read from there, if already present.
    @return: Stimuli, FixationTrains

    .. note::
        This code needs a working matlab or octave installation as the original
        matlab code by Judd et al. is used to extract the fixation from the
        eyetracking data.

        The first fixation of each fixation train is discarded as stated in the
        paper (Judd et al. 2009).

    .. seealso::

        Tilke Judd, Krista Ehinger, Fredo Durand, Antonio Torralba. Learning to Predict where Humans Look [ICCV 2009]

        http://people.csail.mit.edu/tjudd/WherePeopleLook/index.html
    """
    if include_initial_fixation:
        first_fixation = 0
    else:
        first_fixation = 1

    if location:
        location = os.path.join(location, name)
        if os.path.exists(location):
            stimuli = _load(os.path.join(location, 'stimuli.hdf5'))
            fixations = _load(os.path.join(location, 'fixations.hdf5'))
            return stimuli, fixations
        os.makedirs(location)
    with TemporaryDirectory(cleanup=True) as temp_dir:
        download_and_check('http://saliency.mit.edu/trainSet.zip',
                           os.path.join(temp_dir, 'trainSet.zip'),
                           '56ad5c77e6c8f72ed9ef2901628d6e48')

        # Stimuli
        print('Creating stimuli')
        f = zipfile.ZipFile(os.path.join(temp_dir, 'trainSet.zip'))
        f.extractall(temp_dir)

        stimuli_src_location = os.path.join(temp_dir, 'trainSet', 'Stimuli')
        stimuli_target_location = os.path.join(location, 'Stimuli') if location else None
        images = glob.glob(os.path.join(stimuli_src_location, '**', '*.jpg'))
        images = [os.path.relpath(img, start=stimuli_src_location) for img in images]
        stimuli_filenames = natsorted(images)

        stimuli = create_stimuli(stimuli_src_location, stimuli_filenames, stimuli_target_location)

        # FixationTrains

        print('Creating fixations')

        with open(os.path.join(temp_dir, 'load_cat2000.m'), 'wb') as f:
            f.write(resource_string(__name__, 'scripts/{}'.format('load_cat2000.m')))

        cmds = []
        # It is vital _not_ to store the extracted fixations in the main
        # directory where matlab is running, as matlab will check the timestamp
        # of all files in this directory very often. This leads to heavy
        # performance penalties and would make matlab run for more than an
        # hour.
        out_path = 'extracted'
        os.makedirs(os.path.join(temp_dir, out_path))
        run_matlab_cmd('load_cat2000;', cwd=temp_dir)

        print('Extracting fixations. This can take some minutes.')
        print('Warning: In the IPython Notebook, the output is shown on the console instead of the notebook.')

        #run_matlab_cmd('extract_all_fixations;', cwd=temp_dir)
        #
        xs = []
        ys = []
        ts = []
        ns = []
        train_subjects = []
        subject_dict = {}

        files = natsorted(glob.glob(os.path.join(temp_dir, out_path, '*.mat')))
        for f in progressinfo(files):
            mat_data = loadmat(f)
            fix_data = mat_data['data']
            name = mat_data['name'][0]
            n = int(os.path.basename(f).split('fix', 1)[1].split('_')[0]) - 1
            stimulus_size = stimuli.sizes[n]
            _, _, subject = name.split('.eye')[0].split('-')
            if subject not in subject_dict:
                subject_dict[subject] = len(subject_dict)
            subject_id = subject_dict[subject]

            x = []
            y = []
            t = []
            for i in range(first_fixation, fix_data.shape[0]):  # Skip first fixation like Judd does via first_fixation=1
                if fix_data[i, 0] < 0 or fix_data[i, 1] < 0:
                    continue
                if fix_data[i, 0] >= stimulus_size[1] or fix_data[i, 1] >= stimulus_size[0]:
                    continue
                if any(np.isnan(fix_data[i])):  # skip invalid data
                    continue
                x.append(fix_data[i, 0])
                y.append(fix_data[i, 1])
                t.append(len(x))  # Eye Tracker rate = 240Hz
            xs.append(x)
            ys.append(y)
            ts.append(t)
            ns.append(n)
            train_subjects.append(subject_id)
        fixations = FixationTrains.from_fixation_trains(xs, ys, ts, ns, train_subjects)

    if location:
        stimuli.to_hdf5(os.path.join(location, 'stimuli.hdf5'))
        fixations.to_hdf5(os.path.join(location, 'fixations.hdf5'))
    return stimuli, fixations


def get_iSUN(location=None):
    """
    Loads or downloads and caches the iSUN dataset.
    @type  location: string, defaults to `None`
    @param location: If and where to cache the dataset. The dataset
                     will be stored in the subdirectory `iSUN` of
                     location and read from there, if already present.
    @return: Training stimuli, validation stimuli, testing stimuli, training fixation trains, validation fixation trains

    .. seealso::

        P. Xu, K. A. Ehinger, Y. Zhang, A. Finkelstein, S. R. Kulkarni, and J. Xiao.: TurkerGaze: Crowdsourcing Saliency with Webcam based Eye Tracking

        http://lsun.cs.princeton.edu/

        http://vision.princeton.edu/projects/2014/iSUN/
    """
    if location:
        location = os.path.join(location, 'iSUN')
        if os.path.exists(location):
            stimuli_training = _load(os.path.join(location, 'stimuli_training.hdf5'))
            stimuli_validation = _load(os.path.join(location, 'stimuli_validation.hdf5'))
            stimuli_testing = _load(os.path.join(location, 'stimuli_testing.hdf5'))
            fixations_training = _load(os.path.join(location, 'fixations_training.hdf5'))
            fixations_validation = _load(os.path.join(location, 'fixations_validation.hdf5'))
            return stimuli_training, stimuli_validation, stimuli_testing, fixations_training, fixations_validation
        os.makedirs(location)
    with TemporaryDirectory(cleanup=True) as temp_dir:
        download_and_check('http://lsun.cs.princeton.edu/challenge/2015/eyetracking/data/training.mat',
                           os.path.join(temp_dir, 'training.mat'),
                           '5a8b15134b17c7a3f69b087845db1363')
        download_and_check('http://lsun.cs.princeton.edu/challenge/2015/eyetracking/data/validation.mat',
                           os.path.join(temp_dir, 'validation.mat'),
                           'f68e9b011576e48d2460b883854fd86c')
        download_and_check('http://lsun.cs.princeton.edu/challenge/2015/eyetracking/data/testing.mat',
                           os.path.join(temp_dir, 'testing.mat'),
                           'be008ef0330467dcb9c9cd9cc96a8546')
        download_and_check('http://lsun.cs.princeton.edu/challenge/2015/eyetracking/data/fixation.zip',
                           os.path.join(temp_dir, 'fixation.zip'),
                           'aadc15784e1b0023cda4536335b7839c')
        download_and_check('http://lsun.cs.princeton.edu/challenge/2015/eyetracking/data/image.zip',
                           os.path.join(temp_dir, 'image.zip'),
                           '0a3af01c5307f1d44f5dd309f71ea963')

        # Stimuli
        print('Creating stimuli')
        f = zipfile.ZipFile(os.path.join(temp_dir, 'image.zip'))
        namelist = f.namelist()
        namelist = filter_files(namelist, ['.DS_Store'])
        f.extractall(temp_dir, namelist)

        def get_stimuli_names(name):
            data_file = os.path.join(temp_dir, '{}.mat'.format(name))
            data = loadmat(data_file)[name]
            stimuli_names = [d[0] for d in data['image'][:, 0]]
            stimuli_names = ['{}.jpg'.format(n) for n in stimuli_names]
            return stimuli_names

        stimulis = []
        stimuli_src_location = os.path.join(temp_dir, 'images')
        for name in ['training', 'validation', 'testing']:
            print("Creating {} stimuli".format(name))
            stimuli_target_location = os.path.join(location, 'stimuli_{}'.format(name)) if location else None
            images = get_stimuli_names(name)
            stimulis.append(create_stimuli(stimuli_src_location, images, stimuli_target_location))

        # FixationTrains
        print('Creating fixations')

        def get_fixations(name):
            data_file = os.path.join(temp_dir,'{}.mat'.format(name))
            data = loadmat(data_file)[name]
            gaze = data['gaze'][:, 0]
            ns = []
            train_xs = []
            train_ys = []
            train_ts = []
            train_subjects = []
            for n in range(len(gaze)):
                fixation_trains = gaze[n]['fixation'][0, :]
                for train in fixation_trains:
                    xs = train[:, 0]
                    ys = train[:, 1]
                    ns.append(n)
                    train_xs.append(xs)
                    train_ys.append(ys)
                    train_ts.append(range(len(xs)))
                    train_subjects.append(0)
            fixations = FixationTrains.from_fixation_trains(train_xs, train_ys, train_ts, ns, train_subjects)
            return fixations

        fixations = []
        for name in ['training', 'validation']:
            print("Creating {} fixations".format(name))
            fixations.append(get_fixations(name))

    if location:
        stimulis[0].to_hdf5(os.path.join(location, 'stimuli_training.hdf5'))
        stimulis[1].to_hdf5(os.path.join(location, 'stimuli_validation.hdf5'))
        stimulis[2].to_hdf5(os.path.join(location, 'stimuli_test.hdf5'))
        fixations[0].to_hdf5(os.path.join(location, 'fixations_training.hdf5'))
        fixations[1].to_hdf5(os.path.join(location, 'fixations_validation.hdf5'))

    return stimulis + fixations


def get_iSUN_training(location=None):
    """
    @return: Training stimuli, training fixation trains

    See `get_iSUN` for more information"""
    training_stimuli, validation_stimuli, test_stimuli, training_fixations, validation_fixations = get_iSUN(location=location)
    return training_stimuli, training_fixations


def get_iSUN_validation(location=None):
    """
    @return: validation stimuli, validation fixation trains

    See `get_iSUN` for more information"""
    training_stimuli, validation_stimuli, test_stimuli, training_fixations, validation_fixations = get_iSUN(location=location)
    return validation_stimuli, validation_fixations


def get_iSUN_testing(location=None):
    """
    @return: testing stimuli

    See `get_iSUN` for more information"""
    training_stimuli, validation_stimuli, test_stimuli, training_fixations, validation_fixations = get_iSUN(location=location)
    return test_stimuli


def _get_SALICON_obsolete(data_type, stimuli_url, stimuli_hash, fixation_url, fixation_hash, location):
    """Obsolete code for loading from the SALICON json files. The URLs don't exist anymore and the LSUN challenge uses simpler mat files"""
    from salicon.salicon import SALICON
    with TemporaryDirectory(cleanup=True) as temp_dir:
        download_and_check(stimuli_url,
                           os.path.join(temp_dir, 'stimuli.zip'),
                           stimuli_hash)
        if fixation_url is not None:
            download_and_check(fixation_url,
                               os.path.join(temp_dir, 'fixations.json'),
                               fixation_hash)
        # Stimuli

        annFile = os.path.join(temp_dir, 'fixations.json')
        salicon = SALICON(annFile)

        # get all images
        imgIds = salicon.getImgIds()
        images = salicon.loadImgs(imgIds)

        print('Creating stimuli')
        f = zipfile.ZipFile(os.path.join(temp_dir, 'stimuli.zip'))
        namelist = f.namelist()
        f.extractall(temp_dir, namelist)
        del f

        stimuli_src_location = os.path.join(temp_dir, data_type)
        stimuli_target_location = os.path.join(location, 'stimuli') if location else None
        filenames = [img['file_name'] for img in images]
        stimuli = create_stimuli(stimuli_src_location, filenames, stimuli_target_location)

        if fixation_url is not None:
            # FixationTrains
            print('Creating fixations')

            ns = []
            train_xs = []
            train_ys = []
            train_ts = []
            train_subjects = []

            for n, imgId in progressinfo(enumerate(imgIds)):
                annIds = salicon.getAnnIds(imgIds=imgId)
                anns = salicon.loadAnns(annIds)
                for ann in anns:
                    fs = ann['fixations']
                    # SALICON annotations are 1-indexed, not 0-indexed.
                    xs = np.array([f[1]-1 for f in fs])
                    ys = np.array([f[0]-1 for f in fs])
                    ns.append(np.ones(len(xs), dtype=int)*n)
                    train_xs.append(xs)
                    train_ys.append(ys)
                    train_ts.append(range(len(xs)))
                    train_subjects.append(np.ones(len(xs), dtype=int)*ann['worker_id'])

            xs = np.hstack(train_xs)
            ys = np.hstack(train_ys)
            ts = np.hstack(train_ts)

            ns = np.hstack(ns)
            subjects = np.hstack(train_subjects)

            fixations = Fixations.FixationsWithoutHistory(xs, ys, ts, ns, subjects)
        else:
            fixations = None

    if location:
        stimuli.to_hdf5(os.path.join(location, 'stimuli.hdf5'))
        if fixations is not None:
            fixations.to_hdf5(os.path.join(location, 'fixations.hdf5'))
    return stimuli, fixations



def get_SALICON(edition='2015', fixation_type='mouse', location=None):
    """
    Loads or downloads and caches the SALICON dataset as used in the LSUN challenge prior to 2017.
    For memory reasons no fixation trains  are provided.
    @type edition: string, defaults to '2015'
    @param edition: whether to provide the original SALICON dataset from 2015 or the update from 2017

    @type fixation_type: string, defaults to 'mouse'
    @param fixation_type: whether to use the mouse gaze postion ('mouse')
                          or the inferred fixations ('fixations').
                          For more details see the SALICON challenge homepage (below).

    @type  location: string, defaults to `None`
    @param location: If and where to cache the dataset. The dataset
                     will be stored in the subdirectory `SALICON` of
                     location and read from there, if already present.
    @return: Training stimuli, validation stimuli, testing stimuli, training fixation trains, validation fixation trains

    .. seealso::
        Ming Jiang, Shengsheng Huang, Juanyong Duan*, Qi Zhao: SALICON: Saliency in Context, CVPR 2015

        http://salicon.net/

    .. note:
        prior to version 0.2.0 of pysaliency the data was stored in a way that allowed
        accessing the subject ids. This is not possible anymore and each scanpath on
        each image is just assigned an auto-incrementing id.
    """
    if edition not in ['2015', '2017']:
        raise ValueError('edition has to be \'2015\' or \'2017\', not \'{}\''.format(edition))

    if fixation_type not in ['mouse', 'fixations']:
        raise ValueError('fixation_type has to be \'mouse\' or \'fixations\', not \'{}\''.format(fixation_type))

    name = _get_SALICON_name(edition=edition, fixation_type=fixation_type)

    stimuli_train, stimuli_val, stimuli_test = _get_SALICON_stimuli(
        location=location, name='SALICON', edition=edition, fixation_type=fixation_type)
    fixations_train, fixations_val = _get_SALICON_fixations(
        location=location, name=name, edition=edition, fixation_type=fixation_type)

    return stimuli_train, stimuli_val, stimuli_test, fixations_train, fixations_val


def _get_SALICON_name(edition='2015', fixation_type='mouse'):
    if edition not in ['2015', '2017']:
        raise ValueError('edition has to be \'2015\' or \'2017\', not \'{}\''.format(edition))

    if fixation_type not in ['mouse', 'fixations']:
        raise ValueError('fixation_type has to be \'mouse\' or \'fixations\', not \'{}\''.format(fixation_type))

    if edition == '2015':
        if fixation_type == 'mouse':
            name = 'SALICON'
        elif fixation_type == 'fixations':
            name = 'SALICON2015_fixations'
    elif edition == '2017':
        if fixation_type == 'mouse':
            name = 'SALICON2017_mouse'
        elif fixation_type == 'fixations':
            name = 'SALICON2017_fixations'

    return name


def _get_SALICON_stimuli(location, name, edition='2015', fixation_type='mouse'):
    if edition not in ['2015', '2017']:
        raise ValueError('edition has to be \'2015\' or \'2017\', not \'{}\''.format(edition))

    if fixation_type not in ['mouse', 'fixations']:
        raise ValueError('fixation_type has to be \'mouse\' or \'fixations\', not \'{}\''.format(fixation_type))

    if location:
        location = os.path.join(location, name)
        if os.path.exists(location):
            if all(os.path.exists(os.path.join(location, filename)) for filename in ['stimuli_train.hdf5', 'stimuli_val.hdf5', 'stimuli_test.hdf5']):
                stimuli_train = read_hdf5(os.path.join(location, 'stimuli_train.hdf5'))
                stimuli_val = read_hdf5(os.path.join(location, 'stimuli_val.hdf5'))
                stimuli_test = read_hdf5(os.path.join(location, 'stimuli_test.hdf5'))
                return stimuli_train, stimuli_val, stimuli_test
        os.makedirs(location, exist_ok=True)

    with TemporaryDirectory(cleanup=True) as temp_dir:
        stimuli_file = os.path.join(temp_dir, 'stimuli.zip')

        download_and_check(
            'http://lsun.cs.princeton.edu/challenge/2015/eyetracking_salicon/data/image.zip',
            stimuli_file,
            '856e451461b8c164f556ebef96dad1a2'
        )

        print("Extracting stimuli")

        f = zipfile.ZipFile(stimuli_file)
        f.extractall(temp_dir)

        stimuli_train = create_stimuli(
            stimuli_location = os.path.join(temp_dir, 'images'),
            filenames = [os.path.basename(f) for f in sorted(glob.glob(os.path.join(temp_dir, 'images', 'COCO_train*')))],
            location=os.path.join(location, 'stimuli', 'train') if location else None
        )

        stimuli_val = create_stimuli(
            stimuli_location = os.path.join(temp_dir, 'images'),
            filenames = [os.path.basename(f) for f in sorted(glob.glob(os.path.join(temp_dir, 'images', 'COCO_val*')))],
            location=os.path.join(location, 'stimuli', 'val') if location else None
        )

        stimuli_test = create_stimuli(
            stimuli_location = os.path.join(temp_dir, 'images'),
            filenames = [os.path.basename(f) for f in sorted(glob.glob(os.path.join(temp_dir, 'images', 'COCO_test*')))],
            location=os.path.join(location, 'stimuli', 'test') if location else None
        )

        if location is not None:
            stimuli_train.to_hdf5(os.path.join(location, 'stimuli_train.hdf5'))
            stimuli_val.to_hdf5(os.path.join(location, 'stimuli_val.hdf5'))
            stimuli_test.to_hdf5(os.path.join(location, 'stimuli_test.hdf5'))

        return stimuli_train, stimuli_val, stimuli_test


def _get_SALICON_fixations(location, name, edition='2015', fixation_type='mouse'):
    if edition not in ['2015', '2017']:
        raise ValueError('edition has to be \'2015\' or \'2017\', not \'{}\''.format(edition))

    if fixation_type not in ['mouse', 'fixations']:
        raise ValueError('fixation_type has to be \'mouse\' or \'fixations\', not \'{}\''.format(fixation_type))

    if location:
        location = os.path.join(location, name)
        if os.path.exists(location):
            if all(os.path.exists(os.path.join(location, filename)) for filename in ['fixations_train.hdf5', 'fixations_val.hdf5']):
                fixations_train = read_hdf5(os.path.join(location, 'fixations_train.hdf5'))
                fixations_val = read_hdf5(os.path.join(location, 'fixations_val.hdf5'))
                return fixations_train, fixations_val
        os.makedirs(location, exist_ok=True)

    with TemporaryDirectory(cleanup=True) as temp_dir:
        fixations_file = os.path.join(temp_dir, 'fixations.zip')

        if edition == '2015':
            download_file_from_google_drive('0B2hsWbciDVedWHFiMUVVWFRZTE0', fixations_file)
            check_file_hash(fixations_file, '9a22db9d718200fb90252e5010c004c4')
        elif edition == '2017':
            download_file_from_google_drive('0B2hsWbciDVedS1lBZHprdXFoZkU', fixations_file)
            check_file_hash(fixations_file, '462b70f4f9e8ea446ac628e46cea8d3d')

        f = zipfile.ZipFile(fixations_file)
        f.extractall(os.path.join(temp_dir, 'fixations'))

        fixations = []

        if fixation_type == 'mouse':
            fixation_attr = 'location'
        elif fixation_type == 'fixations':
            fixation_attr = 'fixations'

        for dataset in ['train', 'val']:
            ns = []
            train_xs = []
            train_ys = []
            train_ts = []
            train_subjects = []

            subject_id = 0

            data_files = list(sorted(glob.glob(os.path.join(temp_dir, 'fixations', dataset, '*.mat'))))
            for n, filename in enumerate(tqdm(data_files)):
                fixation_data = loadmat(filename)
                for subject_data in fixation_data['gaze'].flatten():
                    train_xs.append(subject_data[fixation_attr][:, 0] - 1)  # matlab: one-based indexing
                    train_ys.append(subject_data[fixation_attr][:, 1] - 1)
                    if fixation_type == 'mouse':
                        train_ts.append(subject_data['timestamp'].flatten())
                    elif fixation_type == 'fixations':
                        train_ts.append(range(len(train_xs[-1])))

                    train_subjects.append(np.ones(len(train_xs[-1]), dtype=int)*subject_id)
                    ns.append(np.ones(len(train_xs[-1]), dtype=int)*n)
                    subject_id += 1

            xs = np.hstack(train_xs)
            ys = np.hstack(train_ys)
            ts = np.hstack(train_ts)
            subjects = np.hstack(train_subjects)
            ns = np.hstack(ns)

            fixations.append(Fixations.FixationsWithoutHistory(xs, ys, ts, ns, subjects))

        fixations_train, fixations_val = fixations

        if location is not None:
            fixations_train.to_hdf5(os.path.join(location, 'fixations_train.hdf5'))
            fixations_val.to_hdf5(os.path.join(location, 'fixations_val.hdf5'))

        return fixations_train, fixations_val


def get_SALICON_train(edition='2015', fixation_type='mouse', location=None):
    """
    Loads or downloads and caches the SALICON training dataset. See `get_SALICON` for more details.
    """
    if location:
        name = _get_SALICON_name(edition=edition, fixation_type=fixation_type)
        if os.path.exists(os.path.join(location, name)):
            stimuli = _load(os.path.join(location, 'SALICON', 'stimuli_train.hdf5'))
            fixations = _load(os.path.join(location, name, 'fixations_train.hdf5'))
            return stimuli, fixations
    stimuli_train, stimuli_val, stimuli_test, fixations_train, fixations_val = get_SALICON(location=location, edition=edition, fixation_type=fixation_type)

    return stimuli_train, fixations_train


def get_SALICON_val(edition='2015', fixation_type='mouse',  location=None):
    """
    Loads or downloads and caches the SALICON validation dataset. See `get_SALICON` for more details.
    """
    if location:
        name = _get_SALICON_name(edition=edition, fixation_type=fixation_type)
        if os.path.exists(os.path.join(location, name)):
            stimuli = _load(os.path.join(location, 'SALICON', 'stimuli_val.hdf5'))
            fixations = _load(os.path.join(location, name, 'fixations_val.hdf5'))
            return stimuli, fixations
    stimuli_train, stimuli_val, stimuli_test, fixations_train, fixations_val = get_SALICON(location=location, edition=edition, fixation_type=fixation_type)

    return stimuli_val, stimuli_val


def get_SALICON_test(edition='2015', fixation_type='mouse', location=None):
    """
    Loads or downloads and caches the SALICON test dataset. See `get_SALICON` for more details.
    """
    if location:
        name = _get_SALICON_name(edition=edition, fixation_type=fixation_type)
        if os.path.exists(os.path.join(location, name)):
            stimuli = _load(os.path.join('SALICON', 'stimuli_test.hdf5'))
            return stimuli
    stimuli_train, stimuli_val, stimuli_test, fixations_train, fixations_val = get_SALICON(location=location, edition=edition, fixation_type=fixation_type)

    return stimuli_test


def _get_koehler_fixations(data, task, n_stimuli):
    tasks = {'freeviewing': 'freeview',
             'objectsearch': 'objsearch',
             'saliencysearch': 'salview'}
    task = tasks[task]

    data_x = data['{}_x'.format(task)]
    data_y = data['{}_y'.format(task)]

    # Load Fixation Data
    xs = []
    ys = []
    ts = []
    ns = []
    subjects = []
    subject_ids = range(data_x.shape[0])

    for n, subject_id in tqdm(list(itertools.product(range(n_stimuli), subject_ids))):
            x = data_x[subject_id, n, :] - 1
            y = data_y[subject_id, n, :] - 1
            inds = np.logical_not(np.isnan(x))
            x = x[inds]
            y = y[inds]
            xs.append(x)
            ys.append(y)
            ts.append(range(len(x)))
            ns.append(n)
            subjects.append(subject_id)
    return FixationTrains.from_fixation_trains(xs, ys, ts, ns, subjects)


def get_koehler(location=None, datafile=None):
    """
    Loads or or extracts and caches the Koehler dataset. The dataset
    consists of 800 color images of outdoor and indoor scenes
    of size 405x405px and the fixations for three different tasks:
    free viewing, object search and saliency search.

    @type  location: string, defaults to `None`
    @param location: If and where to cache the dataset. The dataset
                     will be stored in the subdirectory `koehler` of
                     location and read from there, if already present.
    @return: stimuli, fixations_freeviewing, fixations_objectsearch, fixations_saliencysearch

    .. note:: As this dataset is only after filling a download form, pysaliency
              cannot download it for you. Instead you have to download the file
              `PublicData.zip` and provide it to this function via the `datafile`
              keyword argument on the first call.

    .. seealso::

        Kathryn Koehler, Fei Guo, Sheng Zhang, Miguel P. Eckstein. What Do Saliency Models Predict? [JoV 2014]

        http://www.journalofvision.org/content/14/3/14.full

        https://labs.psych.ucsb.edu/eckstein/miguel/research_pages/saliencydata.html
    """
    if location:
        location = os.path.join(location, 'Koehler')
        if os.path.exists(location):
            stimuli = _load(os.path.join(location, 'stimuli.hdf5'))
            fixations_freeviewing = _load(os.path.join(location, 'fixations_freeviewing.hdf5'))
            fixations_objectsearch = _load(os.path.join(location, 'fixations_objectsearch.hdf5'))
            fixations_saliencysearch = _load(os.path.join(location, 'fixations_saliencysearch.hdf5'))
            return stimuli, fixations_freeviewing, fixations_objectsearch, fixations_saliencysearch
        mkdir_p(location)
    if not datafile:
        raise ValueError('The Koehler dataset is not freely available! You have to '
                         'request the data file from the authors and provide it to '
                         'this function via the datafile argument')
    with TemporaryDirectory() as temp_dir:
        z = zipfile.ZipFile(datafile)
        print('Extracting')
        z.extractall(temp_dir)

        # Stimuli
        stimuli_src_location = os.path.join(temp_dir, 'Images')
        stimuli_target_location = os.path.join(location, 'stimuli') if location else None
        stimuli_filenames = ['image_r_{}.jpg'.format(i) for i in range(1, 801)]

        stimuli = create_stimuli(stimuli_src_location, stimuli_filenames, stimuli_target_location)

        # Fixations

        data = loadmat(os.path.join(temp_dir, 'ObserverData.mat'))

        fs = []

        for task in ['freeviewing', 'objectsearch', 'saliencysearch']:
            fs.append(_get_koehler_fixations(data, task, len(stimuli)))

        if location:
            stimuli.to_hdf5(os.path.join(location, 'stimuli.hdf5'))
            fs[0].to_hdf5(os.path.join(location, 'fixations_freeviewing.hdf5'))
            fs[1].to_hdf5(os.path.join(location, 'fixations_objectsearch.hdf5'))
            fs[2].to_hdf5(os.path.join(location, 'fixations_saliencysearch.hdf5'))
    return [stimuli] + fs


def _load_FIGRIM_data(filename, stimuli_indices, stimulus_type):
    data = loadmat(filename)['allImages'].flatten()
    xs = []
    ys = []
    ts = []
    ns = []
    train_subjects = []
    which_times = []
    which_time_names = ['enc', 'rec', 'rec2']
    stimulus_types = []
    responses = []

    for stimulus_data in data:
        n = stimuli_indices[stimulus_data['filename'][0]]
        category = stimulus_data['category'][0]  # TODO: use
        for subject, subject_data in enumerate(stimulus_data['userdata'].flatten()):
            if not subject_data['trial']:
                # No data for this subject and this stimulus
                continue

            for which_time in which_time_names:
                fixations = subject_data['fixations'][0, 0][which_time]
                if not len(fixations):
                    continue
                #if len(fixations) and which_time != 'enc':
                #    print("Problem:", n, subject_name, which_time)
                subject_response = subject_data['SDT'][0][which_time_names.index(which_time)]

                xs.append(fixations[:, 0])
                ys.append(fixations[:, 1])
                ts.append(np.arange(len(xs[-1])))
                ns.append(n)
                train_subjects.append(subject)

                which_times.append(which_time_names.index(which_time))
                stimulus_types.append(stimulus_type)
                responses.append(subject_response)

    return xs, ys, ts, ns, train_subjects, which_times, stimulus_types, responses


def get_FIGRIM(location=None):
    """
    Loads or downloads and caches the FIGRIM dataset. The dataset
    consists of >2700 scenes of sizes 1000x1000px
    and the fixations of subjects while doing a repetition
    recognition task with 3 seconds presentation time.
    subject responses etc are included.

    @type  location: string, defaults to `None`
    @param location: If and where to cache the dataset. The dataset
                     will be stored in the subdirectory `toronto` of
                     location and read from there, if already present.
    @return: Stimuli, FixationTrains

    .. note::
        This dataset comes with additional annotations:
            - stimulus_type: 0=filler, 1=target
            - which_time: 0=encoding, 1=first recognition, 2=second recognition
            - response: 1=hit, 2=false alarm, 3=miss, 4=correct rejection

    .. seealso::

        Bylinskii, Zoya and Isola, Phillip and Bainbridge, Constance and Torralba, Antonio and Oliva, Aude. Intrinsic and Extrinsic Effects on Image Memorability [Vision research 2015]

        http://figrim.mit.edu/index_eyetracking.html
    """
    if location:
        location = os.path.join(location, 'FIGRIM')
        if os.path.exists(location):
            stimuli = _load(os.path.join(location, 'stimuli.hdf5'))
            fixations = _load(os.path.join(location, 'fixations.hdf5'))
            return stimuli, fixations
        os.makedirs(location)
    with TemporaryDirectory(cleanup=True) as temp_dir:
        download_and_check('http://figrim.mit.edu/Fillers.zip',
                           os.path.join(temp_dir, 'Fillers.zip'),
                           'dc0bc9561b5bc90e158ec32074dd1060')

        download_and_check('http://figrim.mit.edu/Targets.zip',
                           os.path.join(temp_dir, 'Targets.zip'),
                           '2ad3a42ebc377efe4b39064405568201')

        download_and_check('https://github.com/cvzoya/figrim/blob/master/targetData/allImages_release.mat?raw=True',
                           os.path.join(temp_dir, 'allImages_release.mat'),
                           'c72843b05e95ab27594c1d11c849c897')

        download_and_check('https://github.com/cvzoya/figrim/blob/master/fillerData/allImages_fillers.mat?raw=True',
                           os.path.join(temp_dir, 'allImages_fillers.mat'),
                           'ce4f8b4961005d62f7a21191a64cab5e')

        # Stimuli
        mkdir_p(os.path.join(temp_dir, 'stimuli'))
        print('Creating stimuli')
        f = zipfile.ZipFile(os.path.join(temp_dir, 'Fillers.zip'))
        f.extractall(os.path.join(temp_dir, 'stimuli'))

        f = zipfile.ZipFile(os.path.join(temp_dir, 'Targets.zip'))
        f.extractall(os.path.join(temp_dir, 'stimuli'))

        stimuli_src_location = os.path.join(temp_dir, 'stimuli')
        stimuli_target_location = os.path.join(location, 'Stimuli') if location else None
        images = glob.glob(os.path.join(stimuli_src_location, '**', '**', '*.jpg'))
        images = [os.path.relpath(img, start=stimuli_src_location) for img in images]
        stimuli_filenames = natsorted(images)

        stimuli = create_stimuli(stimuli_src_location, stimuli_filenames, stimuli_target_location)

        stimuli_basenames = [os.path.basename(filename) for filename in stimuli_filenames]
        stimulus_indices = {s: stimuli_basenames.index(s) for s in stimuli_basenames}

        # FixationTrains

        print('Creating fixations')

        print('Fillers...')
        (xs_filler,
         ys_filler,
         ts_filler,
         ns_filler,
         train_subjects_filler,
         which_times_filler,
         stimulus_types_filler,
         responses_filler) = _load_FIGRIM_data(os.path.join(temp_dir, 'allImages_fillers.mat'), stimulus_indices, stimulus_type=0)

        print("Targets...")
        (xs_target,
         ys_target,
         ts_target,
         ns_target,
         train_subjects_target,
         which_times_target,
         stimulus_types_target,
         responses_target) = _load_FIGRIM_data(os.path.join(temp_dir, 'allImages_release.mat'), stimulus_indices, stimulus_type=0)

        print("Finalizing...")
        xs = xs_filler + xs_target
        ys = ys_filler + ys_target
        ts = ts_filler + ts_target
        ns = ns_filler + ns_target
        train_subjects = train_subjects_filler + train_subjects_target
        which_times = which_times_filler + which_times_target
        stimulus_types = stimulus_types_filler + stimulus_types_target
        responses = responses_filler + responses_target

        fixations = FixationTrains.from_fixation_trains(
            xs, ys, ts, ns, train_subjects,
            attributes={
                'which_time': which_times,
                'stimulus_type': stimulus_types,
                'response': responses
            })

    if location:
        stimuli.to_hdf5(os.path.join(location, 'stimuli.hdf5'))
        fixations.to_hdf5(os.path.join(location, 'fixations.hdf5'))
    return stimuli, fixations


def get_OSIE(location=None):
    """
    Loads or downloads and caches the OSIE dataset. The dataset
    consists of 700 images of size 800x600px
    and the fixations of 15 subjects while doing a
    freeviewing task with 3 seconds presentation time.

    @type  location: string, defaults to `None`
    @param location: If and where to cache the dataset. The dataset
                     will be stored in the subdirectory `toronto` of
                     location and read from there, if already present.
    @return: Stimuli, FixationTrains

    .. seealso::

        Juan Xu, Ming Jiang, Shuo Wang, Mohan Kankanhalli, Qi Zhao. Predicting Human Gaze Beyond Pixels [JoV 2014]

        http://www-users.cs.umn.edu/~qzhao/predicting.html
    """
    if location:
        location = os.path.join(location, 'OSIE')
        if os.path.exists(location):
            stimuli = _load(os.path.join(location, 'stimuli.hdf5'))
            fixations = _load(os.path.join(location, 'fixations.hdf5'))
            return stimuli, fixations
        os.makedirs(location)
    with TemporaryDirectory(cleanup=True) as temp_dir:
        stimuli_src_location = os.path.join(temp_dir, 'stimuli')
        mkdir_p(stimuli_src_location)
        images = []
        for i in tqdm(list(range(700))):
            filename = '{}.jpg'.format(i+1001)
            target_name = os.path.join(stimuli_src_location, filename)
            urllib.request.urlretrieve(
                'https://github.com/NUS-VIP/predicting-human-gaze-beyond-pixels/raw/master/data/stimuli/'+filename,
                target_name)
            images.append(filename)

        download_and_check('https://github.com/NUS-VIP/predicting-human-gaze-beyond-pixels/blob/master/data/eye/fixations.mat?raw=true',
                           os.path.join(temp_dir, 'fixations.mat'),
                           '8efdf6fe66f38b6e70f854c7ff45aa70')

        # Stimuli
        print('Creating stimuli')

        stimuli_target_location = os.path.join(location, 'Stimuli') if location else None
        stimuli = create_stimuli(stimuli_src_location, images, stimuli_target_location)

        stimulus_indices = {s: images.index(s) for s in images}

        # FixationTrains

        print('Creating fixations')
        data = loadmat(os.path.join(temp_dir, 'fixations.mat'))['fixations'].flatten()

        xs = []
        ys = []
        ts = []
        ns = []
        train_subjects = []

        for stimulus_data in data:
            stimulus_data = stimulus_data[0, 0]
            n = stimulus_indices[stimulus_data['img'][0]]
            for subject, subject_data in enumerate(stimulus_data['subjects'].flatten()):
                fixations = subject_data[0, 0]
                if not len(fixations['fix_x'].flatten()):
                    continue

                xs.append(fixations['fix_x'].flatten())
                ys.append(fixations['fix_y'].flatten())
                ts.append(np.arange(len(xs[-1])))
                ns.append(n)
                train_subjects.append(subject)

    fixations = FixationTrains.from_fixation_trains(xs, ys, ts, ns, train_subjects)

    if location:
        stimuli.to_hdf5(os.path.join(location, 'stimuli.hdf5'))
        fixations.to_hdf5(os.path.join(location, 'fixations.hdf5'))
    return stimuli, fixations


def get_NUSEF_public(location=None):
    """
    Loads or downloads and caches the part of the NUSEF dataset,
    for which the stimuli are public. The dataset
    consists of 758 images of size 1024x700px
    and the fixations of 25 subjects while doing a
    freeviewing task with 5 seconds presentation time.

    Part of the stimuli from NUSEF are available only
    under a special license and only upon request. This
    function returns only the 444 images which are
    available public (and the corresponding fixations).

    @type  location: string, defaults to `None`
    @param location: If and where to cache the dataset. The dataset
                     will be stored in the subdirectory `toronto` of
                     location and read from there, if already present.
    @return: Stimuli, FixationTrains

    .. seealso::

        Subramanian Ramanathan, Harish Katti, Nicu Sebe, Mohan Kankanhalli, Tat-Seng Chua. An eye fixation database for saliency detection in images [ECCV 2010]

        http://mmas.comp.nus.edu.sg/NUSEF.html
    """
    if location:
        location = os.path.join(location, 'NUSEF_public')
        if os.path.exists(location):
            stimuli = _load(os.path.join(location, 'stimuli.hdf5'))
            fixations = _load(os.path.join(location, 'fixations.hdf5'))
            return stimuli, fixations
        os.makedirs(location)
    with TemporaryDirectory(cleanup=True) as temp_dir:

        download_and_check('http://mmas.comp.nus.edu.sg/NUSEF_database.zip',
                           os.path.join(temp_dir, 'NUSEF_database.zip'),
                           '429a78ad92184e8a4b37419988d98953')

        # Stimuli
        print('Creating stimuli')
        f = zipfile.ZipFile(os.path.join(temp_dir, 'NUSEF_database.zip'))
        f.extractall(temp_dir)

        stimuli_src_location = os.path.join(temp_dir, 'NUSEF_database', 'stimuli')
        images = glob.glob(os.path.join(stimuli_src_location, '*.jpg'))
        images = [os.path.relpath(img, start=stimuli_src_location) for img in images]
        stimuli_filenames = sorted(images)

        stimuli_target_location = os.path.join(location, 'Stimuli') if location else None
        stimuli = create_stimuli(stimuli_src_location, stimuli_filenames, stimuli_target_location)

        stimuli_basenames = [os.path.basename(f) for f in stimuli_filenames]
        stimuli_indices = {s: stimuli_basenames.index(s) for s in stimuli_basenames}

        # FixationTrains

        print('Creating fixations')

        xs = []
        ys = []
        ts = []
        ns = []
        train_subjects = []

        scale_x = 1024 / 260
        scale_y = 768 / 280

        fix_location = os.path.join(temp_dir, 'NUSEF_database', 'fix_data')
        for sub_dir in tqdm(os.listdir(fix_location)):
            if not sub_dir+'.jpg' in stimuli_indices:
                # one of the non public images
                continue
            n = stimuli_indices[sub_dir+'.jpg']
            for subject_data in glob.glob(os.path.join(fix_location, sub_dir, '*.fix')):
                data = open(subject_data).read().replace('\r\n', '\n')
                data = data.split('COLS=', 1)[1]
                data = data.split('[Fix Segment Summary')[0]
                lines = data.split('\n')[1:]
                lines = [l for l in lines if l]
                x = []
                y = []
                t = []
                for line in lines:
                    (_,
                     seg_no,
                     fix_no,
                     pln_no,
                     start_time,
                     fix_dur,
                     interfix_dur,
                     interfix_deg,
                     hor_pos,
                     ver_pos,
                     pupil_diam,
                     eye_scn_dist,
                     no_of_flags,
                     fix_loss,
                     interfix_loss) = line.split()
                    x.append(float(hor_pos)*scale_x)
                    y.append(float(ver_pos)*scale_y)
                    t.append(float(start_time.split(':')[-1]))

                xs.append(x)
                ys.append(y)
                ts.append(t)
                ns.append(n)
                train_subjects.append(0)

    fixations = FixationTrains.from_fixation_trains(xs, ys, ts, ns, train_subjects)

    if location:
        stimuli.to_hdf5(os.path.join(location, 'stimuli.hdf5'))
        fixations.to_hdf5(os.path.join(location, 'fixations.hdf5'))
    return stimuli, fixations



from __future__ import absolute_import, division, print_function, unicode_literals

from . import datasets
from . import saliency_map_models
from . import models
from . import external_models
from . import external_datasets

from .datasets import (
    Fixations,
    FixationTrains,
    Stimuli,
    FileStimuli,
    create_nonfixations,
    create_subset,
    remove_out_of_stimulus_fixations,
    concatenate_datasets,
    read_hdf5,
)
from .dataset_config import load_dataset_from_config
from .saliency_map_models import (
    SaliencyMapModel,
    GeneralSaliencyMapModel,
    ScanpathSaliencyMapModel,
    GaussianSaliencyMapModel,
    FixationMap,
    CachedSaliencyMapModel,
    ExpSaliencyMapModel,
    DisjointUnionSaliencyMapModel,
    SubjectDependentSaliencyMapModel,
    ResizingSaliencyMapModel,
    BluringSaliencyMapModel,
    DigitizeMapModel,
    HistogramNormalizedSaliencyMapModel,
    DensitySaliencyMapModel,
    LogDensitySaliencyMapModel,
    EqualizedSaliencyMapModel,
    WTASamplingMixin,
)
from .sampling_models import SamplingModelMixin, ScanpathSamplingModelMixin
from .models import (
    ScanpathModel,
    GeneralModel,
    Model,
    UniformModel,
    CachedModel,
    MixtureModel,
    DisjointUnionModel,
    SubjectDependentModel,
    ShuffledAUCSaliencyMapModel,
    ResizingModel,
    ResizingScanpathModel,
    StimulusDependentModel,
)
from .saliency_map_conversion import (
    SaliencyMapConvertor,
    JointSaliencyMapConvertor,
    optimize_for_information_gain,
)
from .precomputed_models import (
    SaliencyMapModelFromFiles,
    SaliencyMapModelFromDirectory,
    SaliencyMapModelFromFile,
    ModelFromDirectory,
    HDF5SaliencyMapModel,
    HDF5Model,
    export_model_to_hdf5,
)

from .external_models import (
    AIM,
    SUN,
    ContextAwareSaliency,
    BMS,
    GBVS,
    GBVSIttiKoch,
    Judd,
    IttiKoch,
    RARE2012,
    CovSal,
)
from .external_datasets import (
    get_mit1003,
    get_mit1003_onesize,
    get_cat2000_train,
    get_cat2000_test,
    get_toronto,
    get_iSUN_training,
    get_iSUN_validation,
    get_iSUN_testing,
    get_SALICON,
    get_SALICON_train,
    get_SALICON_val,
    get_SALICON_test,
    get_mit300,
    get_koehler,
    get_FIGRIM,
    get_OSIE,
    get_NUSEF_public,
)

from .metric_optimization import maximize_expected_sim, SIMSaliencyMapModel



from __future__ import absolute_import, print_function, division, unicode_literals
import time
import math
import sys
import os, errno


def makedirs(dirname):
    """Creates the directories for dirname via os.makedirs, but does not raise
       an exception if the directory already exists and passes if dirname=""."""
    if not dirname:
        return
    try:
        os.makedirs(dirname)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

def progressinfo(seq, verbose=True, length=None, prefix=''):
	"""Yields from seq while displaying progress information.
	   Unlike mdp.utils.progessinfo, this routine does not
	   display the progress after each iteration but tries
	   to approximate adaequate stepsizes in order to print
	   the progress information roughly ones per second.

       -verbose: if False, the function behaves like `yield from seq`
       -length: can be used to give the length of sequences that
             have no __len__ attribute or to overwrite the length
       -prefix: Will be printed before the status information.
	"""
	if not verbose:
		for item in seq:
			yield item
		return
	next_step = 1
	step_size = 1
	last_time = time.time()
	start_time = last_time
	if length is None:
		if hasattr(seq, '__len__'):
			length = len(seq)
	if length is not None:
		prec = int(math.ceil(math.log10(length)))
		out_string = "\r{prefix}{{count:{prec}d}} ({{ratio:3.1f}}%)".format(prec=prec, prefix=prefix)
	else:
		length = 1
		out_string = "\r{prefix}{{count:d}}".format(prefix=prefix)
	steps = 0
	for i, item in enumerate(seq):
		yield item
		if i == next_step:
			this_time = time.time()
			time_diff = this_time - last_time + 0.0001
			normed_timediff = time_diff / (step_size)
			new_step_size = int(math.ceil(1.0/normed_timediff))
			#In order to avoid overshooting the right stepsize, we take
			#a convex combination with the old stepsize (that will be
			#too small at the beginning)
			step_size = int(math.ceil(0.8*step_size+0.2*new_step_size))
			last_time = this_time
			next_step = i+step_size
			print(out_string.format(count=i, ratio=1.0*i/length*100), end='')
			sys.stdout.flush()
			#steps += 1
	print(out_string.format(count=length, ratio=100.0))
	#end_time = time.time()
	#print "Needed Steps: ", steps
	#print "Last stepsize", step_size
	#print "Needed time", end_time - start_time
	#print "Avg time per step", (end_time - start_time) / steps

def getChunks(seq,verbose=True):
	"""Yields chunks from seq while optionally displaying progress information.
	   after each chunk.
	   This routine tries
	   to approximate adaequate chunksizes in order to print
	   the progress information roughly ones per second.
	"""
	next_step = 1
	step_size = 1
	last_time = time.time()
	start_time = last_time
	length = len(seq)
	prec = int(math.ceil(math.log10(length)))
	out_string = "\r %{0}d (%3.1f %%)".format(prec)
	steps = 0
	next_chunk = []
	for i, item in enumerate(seq):
		next_chunk.append(item)
		if i == next_step:
			yield next_chunk
			next_chunk = []
			this_time = time.time()
			time_diff = this_time - last_time + 0.0001
			normed_timediff = time_diff / (step_size)
			new_step_size = int(math.ceil(1.0/normed_timediff))
			#In order to avoid overshooting the right stepsize, we take
			#a convex combination with the old stepsize (that will be
			#too small at the beginning)
			step_size = int(math.ceil(0.8*step_size+0.2*new_step_size))
			last_time = this_time
			next_step = i+step_size
			if verbose:
				print(out_string % (i, 1.0*i/length*100), end='')
			sys.stdout.flush()
			#steps += 1
	if next_chunk:
		yield next_chunk
	print(out_string % (length, 100.0))
	#end_time = time.time()
	#print "Needed Steps: ", steps
	#print "Last stepsize", step_size
	#print "Needed time", end_time - start_time
	#print "Avg time per step", (end_time - start_time) / steps

def arange_list(l, maxcols=None, empty=False):
    pass

if __name__ == '__main__':
	#new_list= []
	#for chunk in getChunks(range(10000), prefix='test'):
	#	new_list.extend(chunk)
	#assert(new_list == range(10000))

	for i in progressinfo(range(1000), prefix='test'):
		pass
